"""
Tests the DQM compute dispatch module
"""
import copy
import sys
import warnings

import msgpack
import numpy as np
import pytest
import qcelemental as qcel
from qcelemental.tests.test_model_results import center_data

import qcengine as qcng
from qcengine.testing import checkver_and_convert, from_v2, schema_versions, schema_versions2, using, uusing

qcsk_bs = {"name": "custom_basis", "center_data": center_data, "atom_map": ["bs_sto3g_h", "bs_sto3g_h"]}

_canonical_methods = [
    pytest.param("dftd3", {"method": "b3lyp-d3"}, {}, marks=using("classic-dftd3")),
    pytest.param("qcore", {"method": "pbe", "basis": "6-31G"}, {}, marks=using("qcore")),
    pytest.param("molpro", {"method": "hf", "basis": "6-31G"}, {}, marks=using("molpro")),
    pytest.param("mopac", {"method": "PM6"}, {}, marks=using("mopac")),
    pytest.param("mp2d", {"method": "MP2-DMP2"}, {}, marks=using("mp2d")),
    pytest.param("nwchem", {"method": "hf", "basis": "6-31G"}, {}, marks=using("nwchem")),
    pytest.param("openmm", {"method": "openff-1.0.0", "basis": "smirnoff"}, {}, marks=using("openmm")),
    pytest.param("psi4", {"method": "hf", "basis": "6-31G"}, {}, marks=using("psi4")),
    pytest.param("qchem", {"method": "hf", "basis": "6-31G"}, {}, marks=using("qchem")),
    pytest.param("rdkit", {"method": "UFF"}, {}, marks=using("rdkit")),
    pytest.param("terachem_pbs", {"method": "b3lyp", "basis": "6-31G"}, {}, marks=using("terachem_pbs")),
    pytest.param("torchani", {"method": "ANI1x"}, {}, marks=using("torchani")),
    pytest.param("turbomole", {"method": "pbe", "basis": "6-31G"}, {}, marks=using("turbomole")),
    pytest.param("xtb", {"method": "GFN2-xTB"}, {}, marks=using("xtb")),
    pytest.param("adcc", {"method": "adc2", "basis": "6-31G"}, {"n_triplets": 3}, marks=using("adcc")),
    pytest.param("gcp", {"method": "hf3c"}, {}, marks=using("classic-gcp")),
    pytest.param("mrchem", {"method": "blyp"}, {"world_prec": 1.0e-3}, marks=using("mrchem")),
    pytest.param("cfour", {"method": "hf", "basis": "6-31G"}, {}, marks=using("cfour")),
    pytest.param("gamess", {"method": "hf", "basis": "n31"}, {"basis__NGAUSS": 6}, marks=using("gamess")),
    pytest.param("jaguar", {"method": "hf", "basis": "6-31G"}, {"nops": 1}, marks=using("jaguar")),
    pytest.param("mctc-gcp", {"method": "dft/sv"}, {}, marks=using("mctc-gcp")),
    pytest.param("mace", {"method": "small"}, {}, marks=using("mace")),
    pytest.param("aimnet2", {"method": "b973c"}, {}, marks=using("aimnet2")),
    pytest.param("s-dftd3", {"method": "b3lyp-d3"}, {}, marks=using("s-dftd3")),
    pytest.param("dftd4", {"method": "b3lyp-d4"}, {}, marks=using("dftd4")),
    # add as programs available
    # pytest.param("terachem", {"method": "bad"}, marks=using("")),
]

_canonical_methods_qcsk_basis = [
    pytest.param("adcc", {"method": "adc2", "basis": qcsk_bs}, {"n_triplets": 3}, marks=using("adcc")),
    pytest.param("cfour", {"method": "hf", "basis": qcsk_bs}, {}, marks=using("cfour")),
    pytest.param("gamess", {"method": "hf", "basis": qcsk_bs}, {}, marks=using("gamess")),
    pytest.param("jaguar", {"method": "hf", "basis": qcsk_bs}, {"nops": 1}, marks=using("jaguar")),
    pytest.param("molpro", {"method": "hf", "basis": qcsk_bs}, {}, marks=using("molpro")),
    pytest.param("nwchem", {"method": "hf", "basis": qcsk_bs}, {}, marks=using("nwchem")),
    pytest.param("openmm", {"method": "openff-1.0.0", "basis": qcsk_bs}, {}, marks=using("openmm")),
    pytest.param("psi4", {"method": "hf", "basis": qcsk_bs}, {}, marks=using("psi4")),
    pytest.param("qchem", {"method": "hf", "basis": qcsk_bs}, {}, marks=using("qchem")),
    pytest.param("qcore", {"method": "pbe", "basis": qcsk_bs}, {}, marks=using("qcore")),
    pytest.param("turbomole", {"method": "pbe", "basis": qcsk_bs}, {}, marks=using("turbomole")),
]


def _get_molecule(program, molcls):
    if program in ["openmm", "terachem_pbs"]:
        dmol = qcng.get_molecule("water", return_dict=True)
    else:
        dmol = qcng.get_molecule("hydrogen", return_dict=True)
    return molcls(**dmol)


@pytest.mark.parametrize("program, model, keywords", _canonical_methods)
def test_compute_energy(program, model, keywords, schema_versions, request):
    models, retver, _ = schema_versions

    molecule = _get_molecule(program, models.Molecule)

    if from_v2(request.node.name):
        inp = models.AtomicInput(
            molecule=molecule, specification=models.AtomicSpecification(driver="energy", model=model, keywords=keywords)
        )
    else:
        inp = models.AtomicInput(molecule=molecule, driver="energy", model=model, keywords=keywords)

    inp = checkver_and_convert(inp, request.node.name, "pre")
    ret = qcng.compute(inp, program, raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    assert ret.success is True
    assert isinstance(ret.return_result, float)
    assert ret.return_result == ret.properties.return_energy

    # official leave this as dict(), not model_dump(), to ensure remains operational
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        dret = ret.dict()
    assert dret["return_result"] == dret["properties"]["return_energy"]


@pytest.mark.parametrize("program, model, keywords", _canonical_methods)
def test_compute_gradient(program, model, keywords, schema_versions, request):
    models, retver, _ = schema_versions

    molecule = _get_molecule(program, models.Molecule)

    if from_v2(request.node.name):
        inp = models.AtomicInput(
            molecule=molecule,
            specification=models.AtomicSpecification(
                driver="gradient",
                model=model,
                extras={"mytag": "something"},
                keywords=keywords,
            ),
        )
    else:
        inp = models.AtomicInput(
            molecule=molecule,
            driver="gradient",
            model=model,
            extras={"mytag": "something"},
            keywords=keywords,
        )
    if program in ["adcc"]:
        inp = checkver_and_convert(inp, request.node.name, "pre")
        with pytest.raises(qcng.exceptions.InputError) as e:
            qcng.compute(inp, program, raise_error=True, return_version=retver)

        assert "gradient not implemented" in str(e.value)

    else:
        inp = checkver_and_convert(inp, request.node.name, "pre")
        ret = qcng.compute(inp, program, raise_error=True, return_version=retver)
        ret = checkver_and_convert(ret, request.node.name, "post")

        assert ret.success is True
        assert isinstance(ret.return_result, np.ndarray)
        assert len(ret.return_result.shape) == 2
        assert ret.return_result.shape[1] == 3
        if "v2" in request.node.name:
            assert "mytag" in ret.input_data.specification.extras, ret.input_data.specification.extras
            assert "mytag" not in ret.extras, "input extras wrongly present in result"
        else:
            assert "mytag" in ret.extras, ret.extras


@pytest.mark.parametrize("program, model, keywords", _canonical_methods_qcsk_basis)
def test_compute_energy_qcsk_basis(program, model, keywords, schema_versions, request):
    models, retver, _ = schema_versions

    molecule = _get_molecule(program, models.Molecule)
    if from_v2(request.node.name):
        inp = models.AtomicInput(
            molecule=molecule, specification=models.AtomicSpecification(driver="energy", model=model, keywords=keywords)
        )
    else:
        inp = models.AtomicInput(molecule=molecule, driver="energy", model=model, keywords=keywords)

    with pytest.raises(qcng.exceptions.InputError) as e:
        inp = checkver_and_convert(inp, request.node.name, "pre")
        res = qcng.compute(inp, program, raise_error=True, return_version=retver)
        checkver_and_convert(res, request.node.name, "post")

    assert "QCSchema BasisSet for model.basis not implemented" in str(e.value)


@pytest.mark.parametrize(
    "program, model",
    [
        pytest.param("cfour", {"method": "bad"}, marks=using("cfour")),
        pytest.param("dftd3", {"method": "bad"}, marks=using("classic-dftd3")),
        pytest.param("dftd3", {"method": "b3lyp-d3", "driver": "hessian"}, marks=using("classic-dftd3")),
        pytest.param("qcore", {"method": "bad"}, marks=using("qcore")),
        pytest.param("gamess", {"method": "bad"}, marks=using("gamess")),
        pytest.param("mopac", {"method": "bad"}, marks=using("mopac")),
        pytest.param("mp2d", {"method": "bad"}, marks=using("mp2d")),
        pytest.param("nwchem", {"method": "bad"}, marks=using("nwchem")),
        pytest.param("openmm", {"method": "bad"}, marks=using("openmm")),
        pytest.param("psi4", {"method": "bad"}, marks=using("psi4")),
        pytest.param("qchem", {"method": "bad"}, marks=using("qchem")),
        pytest.param("rdkit", {"method": "bad"}, marks=using("rdkit")),
        pytest.param("terachem_pbs", {"method": "bad"}, marks=using("terachem_pbs")),
        pytest.param("torchani", {"method": "bad"}, marks=using("torchani")),
        pytest.param("turbomole", {"method": "bad"}, marks=using("turbomole")),
        pytest.param("adcc", {"method": "bad"}, marks=using("adcc")),
        pytest.param("gcp", {"method": "bad"}, marks=using("classic-gcp")),
        pytest.param("mrchem", {"method": "bad"}, marks=using("mrchem")),
        pytest.param("mctc-gcp", {"method": "bad"}, marks=using("mctc-gcp")),
        pytest.param("mace", {"method": "bad"}, marks=using("mace"))
        # add as programs available
        # ("molpro", {"method": "bad"}),
        # ("terachem", {"method": "bad"}),
        # ("xtb", {"method": "bad"}),
    ],
)
@pytest.mark.parametrize("raiserr", [False, True])
@pytest.mark.parametrize("retdict", [False, True])
@pytest.mark.parametrize("inpdict", [False, True])
def test_compute_bad_models(program, model, schema_versions, request, raiserr, retdict, inpdict):
    models, retver, _ = schema_versions

    amodel = copy.deepcopy(model)
    adriver = amodel.pop("driver", "energy")
    if from_v2(request.node.name):
        inp = {
            "molecule": models.Molecule(**qcng.get_molecule("hydrogen", return_dict=True)),
            "specification": {"model": amodel, "driver": adriver},
        }
    else:
        inp = {
            "molecule": models.Molecule(**qcng.get_molecule("hydrogen", return_dict=True)),
            "model": amodel,
            "driver": adriver,
        }
    if not inpdict:
        inp = models.AtomicInput(**inp)
    inp = checkver_and_convert(inp, request.node.name, "pre")

    if raiserr:
        with pytest.raises(qcng.exceptions.InputError) as exc:
            qcng.compute(inp, program, raise_error=raiserr, return_dict=retdict, return_version=retver)
    else:
        ret = qcng.compute(inp, program, raise_error=raiserr, return_dict=retdict, return_version=retver)
        ret = checkver_and_convert(
            ret, request.node.name, "post", vercheck=False, cast_dict_as="FailedOperation"
        )  # TODO release vercheck=F?
        if retdict:
            assert ret["success"] is False, "wrongly successful"
            assert (
                ret["error"]["error_type"] == "input_error"
            ), f"wrong type: {ret['error']['error_type']=} != 'input_error'"
            if "v2" in request.node.name or "to_v1" in request.node.name:
                assert ret["input_data"]["specification"]["model"]["method"] == model["method"], "input not copied over"
            else:
                assert ret["input_data"]["model"]["method"] == model["method"], "input not copied over"
        else:
            assert ret.success is False, "wrongly successful"
            assert isinstance(
                ret, (qcel.models.v1.FailedOperation, qcel.models.v2.FailedOperation, qcel.models._v1v2.FailedOperation)
            ), "wrong class"
            assert ret.error.error_type == "input_error", f"wrong type: {ret.error.error_type=} != 'input_error'"
            if "v2" in request.node.name:
                ret_method = ret.input_data.specification.model.method
            elif "to_v1" in request.node.name:
                ret_method = ret.input_data["specification"]["model"]["method"]
            else:
                ret_method = ret.input_data["model"]["method"]
            assert ret_method == model["method"], "input not copied over"


@uusing("nwchem")
@pytest.mark.parametrize("envshim", [True, False])
@pytest.mark.parametrize("model", ["Atomic"])
@pytest.mark.parametrize(
    "goodcalc,input_data,return_version,return_dict,xptd",
    [
        # fmt: off
    #           good? in ret  dict?    py313                      py314                      py314+V1V2_SHIM
    pytest.param(True, 1, -1, False,  ("v1.Result",               "v2.FailedOp pyd",         "v2.FailedOp pyd")),          # xptd0
    pytest.param(True, 1,  2, False,  ("v2.Result",               "v2.Result",               "v2.Result")),                # xptd1
    pytest.param(True, 1, -1, True,   ("dict of v1.Result",       "dict of v2.FailedOp pyd", "dict of v1.Result")),        # xptd2
    pytest.param(True, 1,  2, True,   ("dict of v2.Result",       "dict of v2.Result",       "dict of v2.Result")),        # xptd3

    pytest.param(True, 2, -1, False,  ("v2.Result",               "v2.Result",               "v2.Result")),                # xptd4
    pytest.param(True, 2,  1, False , ("v1.Result",               "v2.FailedOp pyd",         "v2.FailedOp pyd")),          # xptd5
    pytest.param(True, 2, -1, True,   ("dict of v2.Result",       "dict of v2.Result",       "dict of v2.Result")),        # xptd6
    pytest.param(True, 2,  1, True,   ("dict of v1.Result",       "dict of v2.FailedOp pyd", "dict of v1.Result")),        # xptd7

    pytest.param(False, 1, -1, False, ("v1.FailedOp mtd",         "v2.FailedOp pyd",         "v2.FailedOp pyd")),          # xptd8
    pytest.param(False, 1,  2, False, ("v2.FailedOp mtd",         "v2.FailedOp mtd",         "v2.FailedOp mtd")),          # xptd9
    pytest.param(False, 1, -1, True,  ("dict of v1.FailedOp mtd", "dict of v2.FailedOp pyd", "dict of v1.FailedOp mtd")),  # xptd10
    pytest.param(False, 1,  2, True,  ("dict of v2.FailedOp mtd", "dict of v2.FailedOp mtd", "dict of v2.FailedOp mtd")),  # xptd11

    pytest.param(False, 2, -1, False, ("v2.FailedOp mtd",         "v2.FailedOp mtd",         "v2.FailedOp mtd")),          # xptd12
    pytest.param(False, 2,  1, False, ("v1.FailedOp mtd",         "v2.FailedOp pyd",         "v2.FailedOp pyd")),          # xptd13
    pytest.param(False, 2, -1, True,  ("dict of v2.FailedOp mtd", "dict of v2.FailedOp mtd", "dict of v2.FailedOp mtd")),  # xptd14
    pytest.param(False, 2,  1, True,  ("dict of v1.FailedOp mtd", "dict of v2.FailedOp pyd", "dict of v1.FailedOp mtd")),
        # xptd15
        # fmt: on
    ],
)
def test_compute_output_table(goodcalc, input_data, return_version, return_dict, xptd, model, envshim, monkeypatch):
    h2 = qcng.get_molecule("hydrogen", return_dict=True)
    method = "hf" if goodcalc else "standard_model_of_particle_physics"
    inp = None

    if model == "Atomic":
        if input_data == 1:
            inp = {"molecule": h2, "driver": "energy", "model": {"method": method, "basis": "sto-3g"}}
        elif input_data == 2:
            inp = {
                "molecule": h2,
                "specification": {"driver": "energy", "model": {"method": method, "basis": "sto-3g"}},
            }
        else:
            pytest.fail(f"Unsupported input_data value for Atomic model: {input_data}")
    else:
        pytest.fail(f"Unsupported model value: {model}")

    if sys.version_info >= (3, 14):
        if envshim:
            monkeypatch.setenv("QCNG_USE_V1V2_SHIM", "1")
            result_this_job = xptd[2]
        else:
            result_this_job = xptd[1]
    else:
        result_this_job = xptd[0]

    ret = qcng.compute(inp, "nwchem", raise_error=False, return_dict=return_dict, return_version=return_version)

    if model == "Atomic":
        # pyd_phrase covers _MSG314 from qcng/compute.py and _MSG2 from qcel/models/v1/__init__.py
        # Append ". You can (a)" to catch the former. use ". Use qcel" to catch the latter
        pyd_phrase = "Reason: pydantic.v1 is unavailable on Python 3.14+"
        mtd_phrase = "Method not recognized"

        if result_this_job == "v2.Result":
            assert isinstance(ret, qcel.models.v2.AtomicResult)

        elif result_this_job == "dict of v2.Result":
            assert isinstance(ret, dict)
            assert ret["schema_name"] == "qcschema_atomic_result"
            assert ret["schema_version"] == 2

        elif result_this_job == "v1.Result":
            assert isinstance(ret, qcel.models.v1.AtomicResult)

        elif result_this_job == "dict of v1.Result":
            assert isinstance(ret, dict)
            assert ret["schema_name"] == "qcschema_output"
            assert ret["schema_version"] == 1

        elif result_this_job.startswith("v2.FailedOp"):
            assert isinstance(ret, qcel.models.v2.FailedOperation)
            errmsg = {"v2.FailedOp mtd": mtd_phrase, "v2.FailedOp pyd": pyd_phrase}[result_this_job]
            assert errmsg in ret.error.error_message

        elif result_this_job.startswith("dict of v2.FailedOp"):
            assert isinstance(ret, dict)
            assert ret["schema_name"] == "qcschema_failed_operation"
            assert ret["schema_version"] == 2
            errmsg = {"dict of v2.FailedOp mtd": mtd_phrase, "dict of v2.FailedOp pyd": pyd_phrase}[result_this_job]
            assert errmsg in ret["error"]["error_message"]

        elif result_this_job.startswith("v1.FailedOp"):
            assert isinstance(ret, qcel.models.v1.FailedOperation)
            errmsg = {"v1.FailedOp mtd": mtd_phrase}[result_this_job]
            assert errmsg in ret.error.error_message

        elif result_this_job.startswith("dict of v1.FailedOp"):
            assert isinstance(ret, dict)
            assert "schema_name" not in ret
            assert "schema_version" not in ret
            errmsg = {"dict of v1.FailedOp mtd": mtd_phrase}[result_this_job]
            assert errmsg in ret["error"]["error_message"]

        else:
            assert 0


@pytest.mark.parametrize(
    "program, model",
    [
        pytest.param("psi4", {"method": "hf", "driver": "eighth"}, marks=using("psi4")),
    ],
)
@pytest.mark.parametrize("raiserr", [False, True])
@pytest.mark.parametrize("retdict", [False, True])
def test_compute_badder_models(program, model, schema_versions2, request, raiserr, retdict):
    models, retver, _ = schema_versions2

    amodel = copy.deepcopy(model)
    adriver = amodel.pop("driver", "energy")
    if from_v2(request.node.name):
        inp = {
            "molecule": models.Molecule(**qcng.get_molecule("hydrogen", return_dict=True)),
            "specification": {"model": amodel, "driver": adriver},
        }
    else:
        inp = {
            "molecule": models.Molecule(**qcng.get_molecule("hydrogen", return_dict=True)),
            "model": amodel,
            "driver": adriver,
        }
    # inp = checkver_and_convert(inp, request.node.name, "pre")

    if raiserr:
        with pytest.raises(qcng.exceptions.InputError) as exc:
            qcng.compute(inp, program, raise_error=raiserr, return_dict=retdict, return_version=retver)
    else:
        ret = qcng.compute(inp, program, raise_error=raiserr, return_dict=retdict, return_version=retver)
        if retdict:
            assert ret["success"] is False, "wrongly successful"
            assert (
                ret["error"]["error_type"] == "input_error"
            ), f"wrong type: {ret['error']['error_type']=} != 'input_error'"
            if from_v2(request.node.name):
                assert ret["input_data"]["specification"]["driver"] == "eighth", "input not copied over"
            else:
                assert ret["input_data"]["driver"] == "eighth", "input not copied over"
        else:
            assert ret.success is False, "wrongly successful"
            assert isinstance(ret, (qcel.models.v1.FailedOperation, qcel.models.v2.FailedOperation)), "wrong class"
            assert ret.error.error_type == "input_error", f"wrong type: {ret.error.error_type=} != 'input_error'"
            # note that input_data *always* a dict in this test (even for v2)
            #   since the error is that the AtomicInput model can't be constructed
            if from_v2(request.node.name):
                assert ret.input_data["specification"]["driver"] == "eighth", "input not copied over"
            else:
                assert ret.input_data["driver"] == "eighth", "input not copied over"


@uusing("psi4")
def test_psi4_restarts(monkeypatch, schema_versions, request):
    """
    Make sure that a random error is raised which can be restarted if psi4 fails with no error message
    """
    models, retver, _ = schema_versions

    # create the psi4 task
    if from_v2(request.node.name):
        inp = models.AtomicInput(
            molecule=models.Molecule(**qcng.get_molecule("hydrogen", return_dict=True)),
            specification={"driver": "energy", "model": {"method": "hf", "basis": "6-31G"}},
        )
    else:
        inp = models.AtomicInput(
            molecule=models.Molecule(**qcng.get_molecule("hydrogen", return_dict=True)),
            driver="energy",
            model={"method": "hf", "basis": "6-31G"},
        )

    def mock_execute(*args, **kwargs):
        """
        Mock the output of a failed psi4 task with missing error message.
        """

        mock_output = {"sucess": False, "outfiles": {"data.msgpack": msgpack.dumps({"missing": "data"})}}
        return True, mock_output

    monkeypatch.setattr("qcengine.programs.psi4.execute", mock_execute)

    inp = checkver_and_convert(inp, request.node.name, "pre")
    with pytest.raises(qcng.exceptions.RandomError):
        _ = qcng.compute(
            input_data=inp, program="psi4", raise_error=True, task_config={"retries": 0}, return_version=retver
        )



import json
import logging
import re
from decimal import Decimal
from typing import TYPE_CHECKING, Optional, Tuple

import numpy as np
import qcelemental as qcel
from qcelemental.molparse import regex

from ..util import PreservingDict

if TYPE_CHECKING:
    from qcelemental.models.v2 import Molecule

logger = logging.getLogger(__name__)


def harvest_output(outtext: str) -> Tuple[PreservingDict, "Molecule", list, str, str]:
    """Function to read an entire NWChem output file.

    Reads all of the different "line search" segments of a file and returns
    values from the last segment for which a geometry was written.

    Args:
        outtext (str): Output written to stdout
    Returns:
        - (PreservingDict) Variables extracted from the output file in the last complete step
        - (Molecule): Molecule from the last complete step
        - (list): Gradient from the last complete step
        - (str): Version string
        - (str): Error message, if any
    """

    # Loop over all steps
    # TODO (wardlt): Is it only necessary to read the last two steps?
    pass_psivar = []
    pass_coord = []
    pass_grad = []
    for outpass in re.split(r" Line search:", outtext, flags=re.MULTILINE):
        psivar, nwcoord, nwgrad, version, module, error = harvest_outfile_pass(outpass)
        pass_psivar.append(psivar)
        pass_coord.append(nwcoord)
        pass_grad.append(nwgrad)

    # Determine which segment contained the last geometry
    retindx = -1 if pass_coord[-1] else -2

    return pass_psivar[retindx], pass_coord[retindx], pass_grad[retindx], version, module, error


def harvest_outfile_pass(outtext):
    """Function to read NWChem output file *outtext* and parse important
    quantum chemical information from it in

    """
    psivar = PreservingDict()
    psivar_coord = None
    psivar_grad = None
    version = ""
    module = None
    error = ""  # TODO (wardlt): The error string is never used.

    NUMBER = r"(?x:" + regex.NUMBER + ")"

    # Process version
    mobj = re.search(
        r"^\s+" + r"Northwest Computational Chemistry Package (NWChem)" + r"\s+" + r"(?:<version>\d+.\d+)" + r"\s*$",
        outtext,
        re.MULTILINE,
    )
    if mobj:
        logger.debug("matched version")
        version = mobj.group("version")

    # Process SCF
    # 1)Fail to converge
    mobj = re.search(r"^\s+" + r"(?:Calculation failed to converge)" + r"\s*$", outtext, re.MULTILINE)
    if mobj:
        logger.debug("failed to converge")

    # 2)Calculation converged
    else:
        mobj = re.search(r"^\s+" + r"(?:Total SCF energy)" + r"\s+=\s*" + NUMBER + r"s*$", outtext, re.MULTILINE)
        if mobj:
            logger.debug("matched HF")
            psivar["HF TOTAL ENERGY"] = mobj.group(1)

        # Process Effective nuclear repulsion energy (a.u.)
        mobj = re.search(
            r"^\s+" + r"Effective nuclear repulsion energy \(a\.u\.\)" + r"\s+" + NUMBER + r"\s*$",
            outtext,
            re.MULTILINE,
        )
        if mobj:
            logger.debug("matched NRE")
            # logger.debug (mobj.group(1))
            psivar["NUCLEAR REPULSION ENERGY"] = mobj.group(1)

        # Process DFT dispersion energy (a.u.)
        mobj = re.search(r"^\s+" + r"(?:Dispersion correction)" + r"\s+=\s*" + NUMBER + r"\s*$", outtext, re.MULTILINE)
        if mobj:
            logger.debug("matched Dispersion")
            logger.debug(mobj.group(1))
            psivar["DISPERSION CORRECTION ENERGY"] = mobj.group(1)

        # Process DFT (RDFT, RODFT,UDFT, SODFT [SODFT for nwchem versions before nwchem 6.8])

        mobj = re.search(r"^\s+" + r"(?:Total DFT energy)" + r"\s+=\s*" + NUMBER + r"\s*$", outtext, re.MULTILINE)
        if mobj:
            logger.debug("matched DFT")
            logger.debug(mobj.group(1))
            psivar["DFT TOTAL ENERGY"] = mobj.group(1)

        # SODFT [for nwchem 6.8+]
        mobj = re.search(
            # fmt: off
            r'^\s+' + r'Total SO-DFT energy' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'Nuclear repulsion energy' + r'\s+' + NUMBER + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        if mobj:
            logger.debug("matched DFT")
            # print (mobj.group(1))
            psivar["DFT TOTAL ENERGY"] = mobj.group(1)
            psivar["NUCLEAR REPULSION ENERGY"] = mobj.group(2)

        # MCSCF
        mobj = re.search(
            # fmt: off
            r'^\s+' + r'Total MCSCF energy' + r'\s+=\s+' + NUMBER + r'\s*$',
            # fmt: off
            outtext,
            re.MULTILINE | re.DOTALL,
        )
        if mobj:
            logger.debug("matched mcscf 2")  # MCSCF energy calculation
            psivar["MCSCF TOTAL ENERGY"] = mobj.group(1)

        mobj = re.findall(
            # fmt: off
            r'^\s+' + r'Total SCF energy' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'One-electron energy' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'Two-electron energy' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'Total MCSCF energy' + r'\s+' + NUMBER + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE | re.DOTALL,
        )

        # for mobj_list in mobj:

        if mobj:  # Need to change to accommodate find all instances
            logger.debug("matched mcscf")  # MCSCF energy calculation
            psivar["HF TOTAL ENERGY"] = mobj.group(1)
            psivar["ONE-ELECTRON ENERGY"] = mobj.group(2)
            psivar["TWO-ELECTRON ENERGY"] = mobj.group(3)
            psivar["MCSCF TOTAL ENERGY"] = mobj.group(4)
        # for mobj_list in mobj:
        #   for i in mobj_list:
        #       count += 0
        # logger.debug('matched mcscf iteration %i', count)
        # psivar['HF TOTAL ENERGY'] = mobj.group(1)
        # psivar['ONE-ELECTRON ENERGY'] = mobj.group(2)
        # psivar['TWO-ELECTRON ENERGY'] = mobj.group(3)
        # psivar['MCSCF TOTAL ENERGY'] = mobj.group(4)

        # Process MP2 (Restricted, Unrestricted(RO n/a))
        # 1)SCF-MP2
        mobj = re.search(
            # fmt: off
            r'^\s+' + r'SCF energy' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'Correlation energy' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'Singlet pairs' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'Triplet pairs' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'Total MP2 energy' + r'\s+' + NUMBER + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE,
        )  # MP2
        if mobj:
            logger.debug("matched scf-mp2")
            module = "mp2grad"
            psivar["HF TOTAL ENERGY"] = mobj.group(1)
            psivar["MP2 CORRELATION ENERGY"] = mobj.group(2)
            psivar["MP2 TOTAL ENERGY"] = mobj.group(5)
        # SCS-MP2
        mobj = re.search(
            # fmt: off
            r'^\s+' + r'Same spin pairs' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'Same spin scaling factor' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'Opposite spin pairs' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'Opposite spin scaling fact.' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'SCS-MP2 correlation energy' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'Total SCS-MP2 energy' + r'\s+' + NUMBER + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        if mobj:
            logger.debug("matched scs-mp2", mobj.groups())
            psivar["MP2 SAME-SPIN CORRELATION ENERGY"] = mobj.group(1)
            psivar["MP2 OPPOSITE-SPIN CORRELATION ENERGY"] = mobj.group(3)

            logger.debug(mobj.group(1))  # ess
            logger.debug(mobj.group(2))  # fss
            logger.debug(mobj.group(3))  # eos
            logger.debug(mobj.group(4))  # fos
            logger.debug(mobj.group(5))  # scs corl
            logger.debug(mobj.group(6))  # scs-mp2

        # 2) DFT-MP2
        mobj = re.search(
            # fmt: off
            r'^\s+' + r'DFT energy' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'Unscaled MP2 energy' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'Total DFT\+MP2 energy' + r'\s+' + NUMBER + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        if mobj:
            logger.debug("matched dft-mp2")
            psivar.pop("DFT TOTAL ENERGY")  # remove previously defined DFT energy w/o DH contribution
            psivar["DFT FUNCTIONAL TOTAL ENERGY"] = mobj.group(1)
            psivar["CURRENT REFERENCE ENERGY"] = mobj.group(1)
            psivar["DFT TOTAL ENERGY"] = mobj.group(3)

        # 3) MP2 with CCSD or CCSD(T) calculation (through CCSD(T) directive)
        mobj = re.search(
            # fmt: off
            r'^\s+' + r'MP2 Energy \(coupled cluster initial guess\)' + r'\s*' +
            r'^\s+' + r'------------------------------------------' + r'\s*' +
            r'^\s+' + r'Reference energy:' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'MP2 Corr\. energy:' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'Total MP2 energy:' + r'\s+' + NUMBER + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        if mobj:
            logger.debug("matched coupled cluster-mp2")
            psivar["MP2 CORRELATION ENERGY"] = mobj.group(2)
            psivar["MP2 TOTAL ENERGY"] = mobj.group(3)

            mobj3 = re.search(r"Final RHF  results", outtext)
            if mobj3:
                psivar["MP2 DOUBLES ENERGY"] = mobj.group(2)

        # 4) Direct MP2

        mobj = re.search(
            # fmt: off
            r'^\s+' + r'SCF energy' + r'\s+' + r"(?P<hf>" + NUMBER + r")" + r'\s*' +
            r'^\s+' + r'Correlation energy' + r'\s+' + r"(?P<mp2corl>" + NUMBER + r")" + r'\s*' +
            r'^\s+' + r'Total MP2 energy' + r'\s+' + r"(?P<mp2>" + NUMBER + r")" + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        mobj2 = re.search(r"Direct MP2", outtext)
        if mobj and mobj2:
            logger.debug("matched direct-mp2")
            module = "directmp2"
            psivar["HF TOTAL ENERGY"] = mobj.group("hf")
            psivar["MP2 CORRELATION ENERGY"] = mobj.group("mp2corl")
            psivar["MP2 TOTAL ENERGY"] = mobj.group("mp2")
            # direct-mp2 is RHF only
            psivar["MP2 DOUBLES ENERGY"] = mobj.group("mp2corl")

        # 5) RI-MP2

        # Process calculation through tce [dertype] command
        tce_cumm_corl = 0.0
        for cc_name in [r"MBPT\(2\)", r"MBPT\(3\)", r"MBPT\(4\)"]:
            mobj = re.search(
                # fmt: off
                r'^\s+' + cc_name + r'\s+' + r'correlation energy / hartree' + r'\s+=\s*' + NUMBER + r'\s*' +
                r'^\s+' + cc_name + r'\s+' + r'total energy / hartree' + r'\s+=\s*' + NUMBER + r'\s*$',
                # fmt: on
                outtext,
                re.MULTILINE,
            )

            mobj3 = re.search(r"Wavefunction type : Restricted open-shell Hartree-Fock", outtext, re.MULTILINE)

            if mobj:
                mbpt_plain = cc_name.replace("\\", "").replace("MBPT", "MP").replace("(", "").replace(")", "")
                logger.debug(f"matched tce mbpt {mbpt_plain}", mobj.groups())
                tce_cumm_corl += float(mobj.group(1))

                if mbpt_plain == "MP2":
                    mobj3 = re.search(r"Wavefunction type : Restricted open-shell Hartree-Fock", outtext, re.MULTILINE)
                    if mobj3:
                        psivar[f"{mbpt_plain} DOUBLES ENERGY"] = mobj.group(1)
                        psivar[f"CURRENT CORRELATION ENERGY"] = mobj.group(1)
                        psivar[f"CURRENT ENERGY"] = Decimal(mobj.group(1)) + psivar[f"HF TOTAL ENERGY"]
                    else:
                        psivar[f"{mbpt_plain} DOUBLES ENERGY"] = mobj.group(1)
                        psivar[f"{mbpt_plain} CORRELATION ENERGY"] = mobj.group(1)
                else:
                    psivar[f"{mbpt_plain} CORRECTION ENERGY"] = mobj.group(1)
                    if not mobj3 and mbpt_plain not in ["MP4"]:
                        psivar[f"{mbpt_plain} DOUBLES ENERGY"] = tce_cumm_corl
                psivar[f"{mbpt_plain} TOTAL ENERGY"] = mobj.group(2)
                module = "tce"

            # TCE dipole- MBPT(n)
            mobj2 = re.search(
                # fmt: off
                r"^\s*" + cc_name + r"\s+" + r"dipole moments \/ hartree & Debye" + r"\s*" +
                r"^\s*" + r"-+" + r"\s*" +
                r'^\s*' + r'X' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
                r'^\s*' + r'Y' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
                r'^\s*' + r'Z' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
                r'^\s*' + r'Total' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*$',
                # fmt: on
                outtext,
                re.MULTILINE,
            )

            if mobj2:
                mbpt_plain = cc_name.replace("\\", "").replace("MBPT", "MP").replace("(", "").replace(")", "")
                logger.debug(f"matched tce {mbpt_plain} dipole moment")
                # only pulling Debye
                dipole = np.array([float(mobj2.group(2)), float(mobj2.group(4)), float(mobj2.group(6))])
                psivar[f"{mbpt_plain} DIPOLE"] = dipole

        # TCE with () or []
        for cc_name in [
            r"CCSD\(T\)",
            r"CCSD\[T\]",
            r"CCSD\(2\)_T",
            r"CCSD\(2\)",
            r"CCSDT\(2\)_Q",
            r"CR-CCSD\[T\]",
            r"CR-CCSD\(T\)",
            r"LR-CCSD\(T\)",
            r"LR-CCSD\(TQ\)-1",
            r"CREOMSD\(T\)",
        ]:
            mobj = re.search(
                # fmt: off
                r'^\s+' + cc_name + r'\s+' + r'correction energy / hartree' + r'\s+=\s*' + NUMBER + r'\s*' +
                r'^\s+' + cc_name + r'\s+' + r'correlation energy / hartree' + r'\s+=\s*' + NUMBER + r'\s*' +
                r'^\s+' + cc_name + r'\s+' + r'total energy / hartree' + r'\s+=\s*' + NUMBER + r'\s*$',
                # fmt: on
                outtext,
                re.MULTILINE,
            )
            if mobj:
                cc_plain = cc_name.replace("\\", "")
                cc_plain.replace("CCSD", "")
                logger.debug(f"matched tce cc {cc_plain}")

                if cc_plain == "CCSD[T]":
                    psivar[f"CCSD+T(CCSD) CORRELATION ENERGY"] = mobj.group(2)
                    psivar[f"CCSD+T(CCSD) TOTAL ENERGY"] = mobj.group(3)
                else:
                    # psivar[f"{cc_corr} CORRECTION ENERGY"] = mobj.group(1)
                    psivar[f"{cc_plain} CORRELATION ENERGY"] = mobj.group(2)
                    psivar[f"{cc_plain} TOTAL ENERGY"] = mobj.group(3)
                module = "tce"

            # TCE dipole with () or []
            mobj2 = re.search(
                # fmt: off
                r"^\s*" + cc_name + r"\s+" + r"dipole moments \/ hartree & Debye" + r"\s*" +
                r"^\s*" + r"-+" + r"\s*" +
                r'^\s*' + r'X' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
                r'^\s*' + r'Y' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
                r'^\s*' + r'Z' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
                r'^\s*' + r'Total' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*$',
                # fmt: on
                outtext,
                re.MULTILINE,
            )

            if mobj2:
                cc_plain = cc_name.replace("\\", "")
                cc_plain.replace("CCSD", "")
                logger.debug(f"matched tce {cc_plain} dipole moment")

                # only pulling Debye
                dipole = np.array([float(mobj2.group(2)), float(mobj2.group(4)), float(mobj2.group(6))])
                psivar[f"{cc_plain} DIPOLE"] = dipole

        # Process other TCE cases
        for cc_name in [
            r"CISD",
            r"QCISD",
            r"CISDT",
            r"CISDTQ",
            r"CCD",
            r"CC2",
            r"CCSD",
            r"CCSDT",
            r"CCSDTQ",
            r"LCCSD",
            r"LCCD",
            r"CCSDTA",
        ]:
            mobj = re.search(
                # fmt: off
                r'^\s+' + r'Iterations converged' + r'\s*' +
                r'^\s+' + cc_name + r'\s+' + r'correlation energy / hartree' + r'\s+=\s*' + NUMBER + r'\s*' +
                r'^\s+' + cc_name + r'\s+' + r'total energy / hartree' + r'\s+=\s*' + NUMBER + r'\s*$',
                # fmt: on
                outtext,
                re.MULTILINE,
            )

            if mobj:
                mobj3 = re.search(r"Wavefunction type : Restricted open-shell Hartree-Fock", outtext, re.MULTILINE)
                logger.debug(f"matched {cc_name}", mobj.groups())
                if mobj3:
                    pass
                else:
                    if cc_name in ["MP2", "MP3", "LCCD", "LCCSD", "CCD", "CCSD"]:
                        psivar[f"{cc_name} DOUBLES ENERGY"] = mobj.group(1)
                psivar[f"{cc_name} CORRELATION ENERGY"] = mobj.group(1)
                psivar[f"{cc_name} TOTAL ENERGY"] = mobj.group(2)
                module = "tce"

            # TCE dipole
            mobj2 = re.search(
                # fmt: off
                r"^\s*" + cc_name + r"\s+" + r"dipole moments \/ hartree & Debye" + r"\s*" +
                r"^\s*" + r"-+" + r"\s*" +
                r'^\s*' + r'X' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
                r'^\s*' + r'Y' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
                r'^\s*' + r'Z' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
                r'^\s*' + r'Total' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*$',
                # fmt: on
                outtext,
                re.MULTILINE,
            )
            if mobj2:
                logger.debug(f"matched tce dipole moment")

                # only pulling Debye
                dipole = np.array([float(mobj2.group(2)), float(mobj2.group(4)), float(mobj2.group(6))])
                psivar[f"{cc_name} DIPOLE"] = dipole
                psivar[f"CURRENT DIPOLE"] = dipole

        # Process CCSD/CCSD(T) using nwchem CCSD/CCSD(T) [dertype] command

        mobj = re.search(
            # fmt: off
            r'^\s+' + r'-----------' + r'\s*' +
            r'^\s+' + r'CCSD Energy' + r'\s*' +
            r'^\s+' + r'-----------' + r'\s*' +
            r'^\s+' + r'Reference energy:' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'CCSD corr\. energy:' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'Total CCSD energy:' + r'\s+' + NUMBER + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE | re.DOTALL,
        )

        if mobj:
            logger.debug("matched ccsd")
            psivar["CCSD CORRELATION ENERGY"] = mobj.group(2)
            psivar["CCSD TOTAL ENERGY"] = mobj.group(3)
            module = "cc"

        mobj = re.search(
            # fmt: off
            r'^\s+' + r'T\(CCSD\) corr\. energy:' + r'\s+' + r"(?P<tccsdcorr>" + NUMBER + r")" + r'\s*' +
            r'^\s+' + r'Total CCSD\+T\(CCSD\) energy:' + r'\s+' + r"(?P<tccsdtot>" + NUMBER + r")" + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE | re.DOTALL,
        )
        if mobj:
            logger.debug("matched ccsd+t(ccsd)")
            psivar["T(CCSD) CORRECTION ENERGY"] = mobj.group("tccsdcorr")
            psivar["CCSD+T(CCSD) CORRELATION ENERGY"] = Decimal(mobj.group("tccsdtot")) - psivar["HF TOTAL ENERGY"]
            psivar["CCSD+T(CCSD) TOTAL ENERGY"] = mobj.group("tccsdtot")
            module = "cc"

        mobj = re.search(
            # fmt: off
            r'^\s+' + r'--------------' + r'\s*' +
            r'^\s+' + r'CCSD\(T\) Energy' + r'\s*' +
            r'^\s+' + r'--------------' + r'\s*' + r'(?:.*?)' +
            r'^\s+' + r'\(T\) corr\. energy:' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'Total CCSD\(T\) energy:' + r'\s+' + NUMBER + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE | re.DOTALL,
        )

        if mobj:
            logger.debug("matched ccsd(t)")
            psivar["(T) CORRECTION ENERGY"] = mobj.group(1)
            psivar["CCSD(T) CORRELATION ENERGY"] = Decimal(mobj.group(2)) - psivar["HF TOTAL ENERGY"]
            psivar["CCSD(T) TOTAL ENERGY"] = mobj.group(2)
            module = "cc"

        mobj = re.search(
            # fmt: off
            r'^\s+' + r'Spin Component Scaled \(SCS\) CCSD' + r'\s*' +
            r'^\s+' + r'-*' + r'\s*' +
            r'^\s+' + r'Same spin contribution:' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'Same spin scaling factor:' + r'\s+' + NUMBER + r'\s*'
            r'^\s+' + r'Opposite spin contribution:' + r'\s+' + NUMBER + r'\s*' +
            #r'^\s+' + r'Opposite spin scaling factor' + r'\s+' + NUMBER + r'\s*'
            r'^\s+' + r'Opposite spin scaling fact.:' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'SCS-CCSD correlation energy:' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'Total SCS-CCSD energy:' + r'\s+' + NUMBER + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE | re.DOTALL,
        )
        # SCS-CCSD included
        if mobj:
            logger.debug("matched scs-ccsd", mobj.groups())
            psivar["CCSD SAME-SPIN CORRELATION ENERGY"] = mobj.group(1)
            psivar["CCSD OPPOSITE-SPIN CORRELATION ENERGY"] = mobj.group(3)
            # psivar['CCSD SAME-SPIN CORRELATION ENERGY'] = psivar['SCS-CCSD SAME-SPIN CORRELATION ENERGY'] = (
            #    Decimal(mobj.group(1)) * Decimal(mobj.group(2)))
            # psivar['CCSD OPPOSITE-SPIN CORRELATION ENERGY'] = psivar['SCS-CCSD OPPOSITE-SPIN CORRELATION ENERGY'] = (
            #    Decimal(mobj.group(4)) * Decimal(mobj.group(3)))
            # psivar['SCS-CCSD CORRELATION ENERGY'] = mobj.group(5)
            # psivar['SCS-CCSD TOTAL ENERGY'] = mobj.group(6)
            # psivar['CUSTOM SCS-CCSD CORRELATION ENERGY'] = 0.5 * (float(
            #    psivar['CCSD SAME-SPIN CORRELATION ENERGY']) + float(psivar['CCSD OPPOSITE-SPIN CORRELATION ENERGY']))
            # psivar['CUSTOM SCS-CCSD TOTAL ENERGY'] = float(mobj.group(6)) + float(
            #   psivar['CUSTOM SCS-CCSD CORRERLATION ENERGY'])

        # Process EOM-[cc_name] #nwchem_tce_dipole = false
        # Parsed information: each symmetry, root excitation energy in eV and total energy in hartree
        # psivar name might need to be fixed
        # each root excitation energy is extracted from the last iteration of right hand side
        mobj = re.findall(
            # fmt: off
            r'^\s+(?:Excited-state calculation \( )(.*)\s+(?:symmetry\))\s+(?:.*\n)*^\s+EOM-' + cc_name +
            # (..) captures symmetry
            r'right-hand side iterations\s+(?:.*\n)*(?:Excited state root)\s+' + NUMBER + #root
            r'\s*(?:Excitation energy / hartree)\s+.\s+' + NUMBER + #excitation energy hartree
            r'\s*(?:/ eV)\s+.\s+' + NUMBER + r'\s*$',
            # excitation energy eV
            # fmt: on
            outtext,
            re.MULTILINE | re.DOTALL,
        )
        # regex should be more dynamic in finding values, need to revisit
        # mobj.group(0) = symmetry value
        # mobj.group(1) = cc_name
        # mobj.group(2) = root number
        # mobj.group(3) = excitation energy (hartree)
        # mobj.group(4) = excitation energy (eV)

        if mobj:
            logger.debug(mobj)
            ext_energy = {}  # dic

            ext_energy_list = []
            logger.debug(f"matched eom-{cc_name}")
            for mobj_list in mobj:
                logger.debug("matched EOM-%s - %s symmetry" % (cc_name, mobj_list[0]))  # cc_name, symmetry
                logger.debug(mobj_list)
                count = 0
                for line in mobj_list[1].splitlines():
                    lline = line.split()
                    logger.debug(lline[1])  # in hartree
                    logger.debug(lline[2])  # in eV
                    count += 1

                    logger.debug("matched excitation energy #%d - %s symmetry" % (count, mobj_list[0]))

                    ext_energy_list.append(lline[1])  # Collect all excitation energies

                    sym = str(mobj_list[0])
                    ext_energy.setdefault(sym, [])
                    ext_energy[sym].append(lline[1])  # Dictionary: symmetries(key), energies(value)

            ext_energy_list.sort(key=float)

            for nroot in range(len(ext_energy_list)):
                for k, e_val in ext_energy.items():
                    if ext_energy_list[nroot] in e_val:
                        symm = k
                        # in hartree
                        psivar[
                            f"EOM-{cc_name} ROOT 0 -> ROOT {nroot + 1} EXCITATION ENERGY - {symm} SYMMETRY"
                        ] = ext_energy_list[nroot]
                        psivar[f"EOM-{cc_name} ROOT 0 -> ROOT {nroot + 1} TOTAL ENERGY - {symm} SYMMETRY"] = psivar[
                            f"{cc_name} TOTAL ENERGY"
                        ] + Decimal(ext_energy_list[nroot])
        gssym = ""
        gs = re.search(r"^\s+" + r"Ground-state symmetry is" + gssym + r"\s*$", outtext, re.MULTILINE)

        if gs:
            logger.debug("matched ground-state symmetry")
            psivar["GROUND-STATE SYMMETRY"] = gssym.group(1)

        # Process TDDFT
        #       1) Spin allowed
        mobj = re.findall(
            # fmt: off
            r'^\s+(?:Root)\s+(\d+)\s+(.*?)\s+' + NUMBER + r'\s+(?:a\.u\.)\s+' + NUMBER + r"\s+eV\s*" +
            r"^\s+" + r"<S2>\s+=\s+" + NUMBER + r"\s*"
            #Root | symmetry | a.u. | eV
            # unkn units for dip/quad
            + r'\s+(?:.*\n)\s+Transition Moments\s+X\s+'+ NUMBER + r'\s+Y\s+'+ NUMBER+ r'\s+Z\s+'+ NUMBER #dipole
            + r'\s+Transition Moments\s+XX\s+'+ NUMBER + r'\s+XY\s+'+ NUMBER+ r'\s+XZ\s+'+ NUMBER #quadrople
            + r'\s+Transition Moments\s+YY\s+'+ NUMBER + r'\s+YZ\s+'+ NUMBER+ r'\s+ZZ\s+'+ NUMBER #quadrople
            + r"\s+" + r"Dipole Oscillator Strength" + r"\s+" + NUMBER + r"\s*$",
            # fmt: on
            outtext,
            re.MULTILINE,
        )

        if mobj:
            logger.debug("matched TDDFT with transition moments")
            for mobj_list in mobj:
                logger.debug(mobj_list)
                iroot = mobj_list[0]
                sym = mobj_list[1]

                # in eV
                psivar[f"TDDFT ROOT {iroot} EXCITATION ENERGY - {sym} SYMMETRY"] = mobj_list[2]
                psivar[f"TDDFT ROOT {iroot} EXCITED STATE ENERGY - {sym} SYMMETRY"] = psivar[
                    "DFT TOTAL ENERGY"
                ] + Decimal(mobj_list[2])
                psivar[f"TDDFT ROOT 0 -> ROOT {iroot} DIPOLE"] = [
                    float(mobj_list[5]),
                    float(mobj_list[6]),
                    float(mobj_list[7]),
                ]
                psivar[f"TDDFT ROOT 0 -> ROOT {iroot} QUADRUPOLE"] = [
                    float(mobj_list[8]),
                    float(mobj_list[9]),
                    float(mobj_list[10]),
                    float(mobj_list[9]),
                    float(mobj_list[11]),
                    float(mobj_list[12]),
                    float(mobj_list[10]),
                    float(mobj_list[12]),
                    float(mobj_list[13]),
                ]
                psivar[f"TDDFT ROOT 0 -> ROOT {iroot} OSCILLATOR STRENGTH (LEN)"] = mobj_list[14]

        #       2) Spin forbidden
        mobj = re.findall(
            # fmt: off
            r'^\s+(?:Root)\s+(\d+)\s+(.*?)\s+' + NUMBER + r'\s(?:a\.u\.)\s+' + NUMBER + r'\s+(?:\w+)'  # Root | symmetry | a.u. | eV
            + r'\s+(?:.\w+.\s+.\s+\d+.\d+)'  # s2 value
            + r'\s+Transition Moments\s+(?:Spin forbidden)' + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        # mobj.group(0) = Root
        # mobj.group(1) = symmetry
        # mobj.group(2) a.u.
        # mobj.group(3) e.V
        # mobj.group(4) Excitation energy
        # mobj.group(5) Excited state energy

        if mobj:
            logger.debug("matched TDDFT - spin forbidden")
            for mobj_list in mobj:
                #### temporary psivars ####
                # in eV
                psivar[f"TDDFT ROOT {mobj_list[0]} EXCITATION ENERGY - {mobj_list[2]} SYMMETRY"] = mobj_list[4]
                psivar[f"TDDFT ROOT {mobj_list[0]} EXCITED STATE ENERGY - {mobj_list[2]} SYMMETRY"] = psivar[
                    "DFT TOTAL ENERGY"
                ] + qcel.constants.converstion_factor("eV", "hartree") * Decimal(mobj_list[4])

                # psivar['TDDFT ROOT %s %s %s EXCITATION ENERGY' %
                #       (mobj_list[0], mobj_list[1], mobj_list[2])] = mobj_list[3]  # in a.u.
                # psivar['TDDFT ROOT %s %s %s EXCITED STATE ENERGY' %(mobj_list[0], mobj_list[1], mobj_list[2])] = \
                #    psivar['DFT TOTAL ENERGY'] + Decimal(mobj_list[3])
            if mobj:
                logger.debug("Non-variation initial energy")  # prints out energy, 5 counts

        # Process geometry
        # 1) CHARGE
        # Read charge from SCF module
        mobj = re.search(
            r"^\s+" + r"charge          =" + r"\s+" + NUMBER + r"\s*$", outtext, re.MULTILINE | re.IGNORECASE
        )

        if mobj:
            logger.debug("matched charge")
            out_charge = int(float(mobj.group(1)))

        # Read charge from General information (not scf module)
        mobj = re.search(
            r"^\s+" + r"Charge           :" + r"\s+" + r"(-?\d+)" + r"\s*$", outtext, re.MULTILINE | re.IGNORECASE
        )

        if mobj:
            logger.debug("matched charge")
            out_charge = int(float(mobj.group(1)))

        # 2) MULTIPLICITY
        # Read multiplicity from SCF module
        mobj = re.search(
            r"^\s+" + r"open shells     =" + r"\s+" + r"(\d+)" + r"\s*$", outtext, re.MULTILINE | re.IGNORECASE
        )

        calcinfo = False

        if mobj:
            logger.debug("matched multiplicity")
            out_mult = int(mobj.group(1)) + 1

        # Read multiplicity from SCF module through alpha, beta electrons
        mobj = re.search(
            # fmt: off
            r'^\s+' + r'alpha electrons =' + r'\s+' + r'(\d+)' + r'\s*' +
            r'^\s+' + r'beta  electrons =' + r'\s+' + r'(\d+)' + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE | re.IGNORECASE,
        )
        if mobj:
            logger.debug("matched multiplicity via alpha and beta electrons 0")
            out_mult = int(mobj.group(1)) - int(mobj.group(2)) + 1  # nopen + 1
            psivar["N ALPHA ELECTRONS"] = mobj.group(1)
            psivar["N BETA ELECTRONS"] = mobj.group(2)

        mobj = re.search(
            # fmt: off
            r"^\s+" + r"Basis functions"       + r"\s+=\s+" + r"(?P<nbf>\d+)" + r"\s*" +
            r"^\s+" + r"Molecular orbitals"    + r"\s+=\s+" + r"(?P<nmo>\d+)" + r"\s*" +
            r"^\s+" + r"Frozen core"           + r"\s+=\s+" + r"(?P<nfc>\d+)" + r"\s*" +
            r"^\s+" + r"Frozen virtuals"       + r"\s+=\s+" + r"(?P<nfv>\d+)" + r"\s*" +
            r"^\s+" + r"Active alpha occupied" + r"\s+=\s+" + r"(?P<nao>\d+)" + r"\s*" +
            r"^\s+" + r"Active beta occupied"  + r"\s+=\s+" + r"(?P<nbo>\d+)" + r"\s*" +
            r"^\s+" + r"Active alpha virtual"  + r"\s+=\s+" + r"(?P<nav>\d+)" + r"\s*" +
            r"^\s+" + r"Active beta virtual"   + r"\s+=\s+" + r"(?P<nbv>\d+)" + r"\s*",
            # fmt: on
            outtext,
            re.MULTILINE | re.IGNORECASE,
        )
        if mobj:
            logger.debug("matched alpha and beta electrons 1", mobj.groups())
            calcinfo = True
            psivar["N BASIS FUNCTIONS"] = mobj.group("nbf")
            psivar["N MOLECULAR ORBITALS"] = mobj.group("nmo")
            psivar["N ALPHA ELECTRONS"] = int(mobj.group("nao")) + int(mobj.group("nfc"))
            psivar["N BETA ELECTRONS"] = int(mobj.group("nbo")) + int(mobj.group("nfc"))

        mobj = re.search(
            # fmt: off
            r"^\s+" + "No. of electrons" + r"\s+:\s+" + r"(?P<ne>\d+)" + r"\s*" +
            r"^\s+" + "Alpha electrons" + r"\s+:\s+" + r"(?P<nae>\d+)" + r"\s*" +
            r"^\s+" + "Beta electrons" + r"\s+:\s+" + r"(?P<nbe>\d+)" + r"\s*" +
            r"^\s+" + "No. of orbitals" + r"\s+:\s+" + r"(?P<no>\d+)" + r"\s*" +
            r"^\s+" + "Alpha orbitals" + r"\s+:\s+" + r"(?P<namo>\d+)" + r"\s*" +
            r"^\s+" + "Beta orbitals" + r"\s+:\s+" + r"(?P<nbmo>\d+)" + r"\s*" +
            r"^\s+" + "Alpha frozen cores" + r"\s+:\s+" + r"(?P<nafc>\d+)" + r"\s*" +
            r"^\s+" + "Beta frozen cores" + r"\s+:\s+" + r"(?P<nbfc>\d+)" + r"\s*" +
            r"^\s+" + "Alpha frozen virtuals" + r"\s+:\s+" + r"(?P<nafv>\d+)" + r"\s*" +
            r"^\s+" + "Beta frozen virtuals" + r"\s+:\s+" + r"(?P<nbfv>\d+)" + r"\s*" +
            r"^\s+" + "Spin multiplicity" + r"\s+:\s+\w+" + r"\s*" +
            r"^\s+" + "Number of AO functions" + r"\s+:\s+" + r"(?P<nbf>\d+)" + r"\s*",
            # fmt: on
            outtext,
            re.MULTILINE | re.IGNORECASE,
        )
        if mobj and not calcinfo:
            logger.debug("matched alpha and beta electrons 2", mobj.groups())
            calcinfo = True
            psivar["N BASIS FUNCTIONS"] = mobj.group("nbf")
            psivar["N MOLECULAR ORBITALS"] = (int(mobj.group("namo")) + int(mobj.group("nbmo"))) / 2
            psivar["N ALPHA ELECTRONS"] = mobj.group("nae")
            psivar["N BETA ELECTRONS"] = mobj.group("nbe")

        mobj = re.search(
            # fmt: off
            r"^\s+" + "functions" + r"\s+=\s+" + r"(?P<nbf>\d+)" + r"\s*" +
            r"^\s+" + "atoms" + r"\s+=\s+" + r"(?P<nat>\d+)" + r"\s*" +
            r"^\s+" + "alpha electrons" + r"\s+=\s+" + r"(?P<nae>\d+)" + r"\s*" +
            r"^\s+" + "beta  electrons" + r"\s+=\s+" + r"(?P<nbe>\d+)" + r"\s*",
            # fmt: on
            outtext,
            re.MULTILINE | re.IGNORECASE,
        )
        if mobj and not calcinfo:
            logger.debug("matched alpha and beta electrons 3", mobj.groups())
            calcinfo = True
            psivar["N BASIS FUNCTIONS"] = mobj.group("nbf")
            psivar["N MOLECULAR ORBITALS"] = mobj.group("nbf")
            psivar["N ALPHA ELECTRONS"] = mobj.group("nae")
            psivar["N BETA ELECTRONS"] = mobj.group("nbe")

        mobj = re.search(
            # fmt: off
            r"^\s+" + "functions"     + r"\s+=\s+" + r"(?P<nbf>\d+)" + r"\s*" +
            r"^\s+" + "atoms"         + r"\s+=\s+" + r"(?P<nat>\d+)" + r"\s*" +
            r"^\s+" + "closed shells" + r"\s+=\s+" + r"(?P<ncl>\d+)" + r"\s*" +
            r"^\s+" + "open shells"   + r"\s+=\s+" + r"(?P<nop>\d+)" + r"\s*",
            # fmt: on
            outtext,
            re.MULTILINE | re.IGNORECASE,
        )
        if mobj and not calcinfo:
            logger.debug("matched alpha and beta electrons 4", mobj.groups())
            calcinfo = True
            psivar["N BASIS FUNCTIONS"] = mobj.group("nbf")
            psivar["N MOLECULAR ORBITALS"] = mobj.group("nbf")  # BAD! TODO
            psivar["N ALPHA ELECTRONS"] = int(mobj.group("ncl")) + int(mobj.group("nop"))
            psivar["N BETA ELECTRONS"] = mobj.group("ncl")

        # Read multiplicity from General information (not scf module)
        mobj = re.search(
            r"^\s+" + r"Spin multiplicity:" + r"\s+" + r"(\d+)" + r"\s*$", outtext, re.MULTILINE | re.IGNORECASE
        )

        if mobj:
            logger.debug("matched multiplicity")
            out_mult = int(mobj.group(1))

        # 3) Initial geometry
        mobj = re.search(
            # fmt: off
            r'^\s+' + r'Geometry' + r'.*' + r'\s*' +
            r'^\s+' + r'(?:-+)\s*' + r'\s+' + r'\n' +
            r'^\s' + r'Output coordinates in ' + r'(.*?)' + r'\s' + r'\(scale by' + r'.*' + r'\s' + r'to convert to a\.u\.\)' + r'\s+' + r'\n' +
            r'^\s+' + r'No\.\       Tag          Charge          X              Y              Z' + r'\s*' +
            r'^\s+' + r'---- ---------------- ---------- -------------- -------------- --------------' + r'\s*' +
            r'((?:\s+([1-9][0-9]*)+\s+(bq)?([A-Z]{1,3})((_\w+)|(\d+))?\s+\d+\.\d+\s+[-+]?\d+\.\d+\s+[-+]?\d+\.\d+\s+[-+]?\d+\.\d+\s*\n)+)' + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE | re.IGNORECASE,
        )

        if mobj:
            logger.debug("matched geom")

            # dinky molecule w/ charge and multiplicity
            if mobj.group(1) == "angstroms":
                molxyz = "%d \n%d %d tag\n" % (len(mobj.group(2).splitlines()), out_charge, out_mult)  # unit = angstrom
                for line in mobj.group(2).splitlines():
                    lline = line.split()
                    molxyz += "%s %16s %16s %16s\n" % (lline[-5], lline[-3], lline[-2], lline[-1])
                    # Jiyoung was collecting charge (-4)? see if this is ok for ghosts
                    # Tag    ,    X,        Y,        Z
                psivar_coord = qcel.models.v2.Molecule(
                    validate=False,
                    **qcel.molparse.to_schema(
                        qcel.molparse.from_string(molxyz, dtype="xyz+", fix_com=True, fix_orientation=True)["qm"],
                        dtype=3,
                    ),
                )

            else:  # unit = a.u.
                molxyz = "%d au\n%d %d tag\n" % (len(mobj.group(2).splitlines()), out_charge, out_mult)
                for line in mobj.group(2).splitlines():
                    lline = line.split()
                    chg_on_center = int(float(lline[-4]))
                    if chg_on_center > 0:
                        tag = f"{lline[-5]}"
                    else:
                        tag = lline[-5].strip()
                        if tag.startswith("bq"):
                            tag = tag[2:]
                        tag = f"@{tag}"
                    molxyz += "%s %16s %16s %16s\n" % (tag, lline[-3], lline[-2], lline[-1])
                    # Tag    ,    X,        Y,        Z
                psivar_coord = qcel.models.v2.Molecule(
                    validate=False,
                    **qcel.molparse.to_schema(
                        qcel.molparse.from_string(molxyz, dtype="xyz+", fix_com=True, fix_orientation=True)["qm"],
                        dtype=3,
                    ),
                )

        # Process gradient
        mobj = re.search(
            # fmt: off
            r'^\s+' + r'.*' + r'ENERGY GRADIENTS' + r'\s*' + r'\s+' + r'\n' +
            r'^\s+' + r'atom               coordinates                        gradient' + r'\s*' +
            r'^\s+' + r'x          y          z           x          y          z' + r'\s*' +
            r'((?:\s+([1-9][0-9]*)+\s+([A-Z][a-x]*)(\d+)?\s+[-+]?\d+\.\d+\s+[-+]?\d+\.\d+\s+[-+]?\d+\.\d+\s+[-+]?\d+\.\d+\s+[-+]?\d+\.\d+\s+[-+]?\d+\.\d+\s*\n)+)' + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE,
        )

        if mobj:
            logger.debug("matched molgrad")
            atoms = []
            psivar_grad = []
            for line in mobj.group(1).splitlines():
                lline = line.split()  # Num, Tag, coord x, coord y, coord z, grad x, grad y, grad z
                # print (lline)
                if lline == []:
                    pass
                else:
                    atoms.append(lline[1])  # Tag
                    psivar_grad.append([float(lline[-3]), float(lline[-2]), float(lline[-1])])
            psivar_grad = np.array(psivar_grad).reshape((-1, 3))

        # Process dipole (Properties)
        mobj = re.search(
            # fmt: off
            r'^\s+' + r'Dipole moment' + r'\s+' + NUMBER + r'\s+' + r'A\.U\.' + r'\s*' +
            r'^\s+' + r'DMX' + r'\s+' + NUMBER + r'.*' +
            r'^\s+' + r'DMY' + r'\s+' + NUMBER + r'.*' +
            r'^\s+' + r'DMZ' + r'\s+' + NUMBER + r'.*' +
            r'^\s+' + r'.*' +
            r'^\s+' + r'Total dipole' + r'\s+' + NUMBER + r'\s+' + r'A\.U\.' + r'\s*' +
            r'^\s+' + r'Dipole moment' + r'\s+' + NUMBER + r'\s' + r'Debye\(s\)' + r'\s*' +
            r'^\s+' + r'DMX' + r'\s+' + NUMBER + r'.*' +
            r'^\s+' + r'DMY' + r'\s+' + NUMBER + r'.*' +
            r'^\s+' + r'DMZ' + r'\s+' + NUMBER + r'.*' +
            r'^\s+' + r'.*' +
            r'^\s+' + r'Total dipole' + r'\s+' + NUMBER + r'\s' + r'DEBYE\(S\)' + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE,
        )

        if mobj:
            logger.debug("matched total dipole")

            # UNIT = DEBYE(S)
            d2au = Decimal(qcel.constants.conversion_factor("debye", "e * bohr"))
            psivar[f"CURRENT DIPOLE"] = d2au * np.array([mobj.group(7), mobj.group(8), mobj.group(9)])
            # total?

            # Process error code
            mobj = re.search(
                # fmt: off
                r'^\s+' + r'current input line \:' + r'\s*' +
                r'^\s+' + r'([1-9][0-9]*)' + r'\:' + r'\s+' + r'(.*)' + r'\s*' +
                r'^\s+' r'------------------------------------------------------------------------' + r'\s*' +
                r'^\s+' r'------------------------------------------------------------------------' + r'\s*' +
                r'^\s+' + r'There is an error in the input file' + r'\s*$',
                # fmt: on
                outtext,
                re.MULTILINE,
            )
            if mobj:
                logger.debug("matched error")
            # print (mobj.group(1)) #error line number
            # print (mobj.group(2)) #error reason
            psivar["NWCHEM ERROR CODE"] = mobj.group(1)
            # TODO process errors into error var

    # Get the size of the basis sets, etc
    mobj = re.search(r"No. of atoms\s+:\s+(\d+)", outtext, re.MULTILINE)
    if mobj:
        psivar["N ATOMS"] = mobj.group(1)
    mobj = re.search(
        r"No. of electrons\s+:\s+(\d+)\s+Alpha electrons\s+:\s+(\d+)\s+Beta electrons\s+:\s+(\d+)",
        outtext,
        re.MULTILINE,
    )
    if mobj:
        psivar["N ALPHA ELECTRONS"] = mobj.group(2)
        psivar["N BETA ELECTRONS"] = mobj.group(3)
        if psivar["N ALPHA ELECTRONS"] == psivar["N BETA ELECTRONS"]:
            # get HOMO and LUMO energy
            mobj = re.search(
                r"Vector"
                + r"\s+"
                + r"%d" % (psivar["N ALPHA ELECTRONS"])
                + r"\s+"
                + r"Occ="
                + r".*"
                + r"\s+"
                + r"E="
                + r"([+-]?\s?\d+[.]\d+)"
                + r"[D]"
                + r"([+-]0\d)",
                outtext,
                re.MULTILINE,
            )
            if mobj:
                homo = float(mobj.group(1)) * (10 ** (int(mobj.group(2))))
                psivar["HOMO"] = np.array([round(homo, 10)])
            mobj = re.search(
                r"Vector"
                + r"\s+"
                + r"%d" % (psivar["N ALPHA ELECTRONS"] + 1)
                + r"\s+"
                + r"Occ="
                + r".*"
                + r"\s+"
                + r"E="
                + r"([+-]?\s?\d+[.]\d+)"
                + r"[D]"
                + r"([+-]0\d)",
                outtext,
                re.MULTILINE,
            )
            if mobj:
                lumo = float(mobj.group(1)) * (10 ** (int(mobj.group(2))))
                psivar["LUMO"] = np.array([round(lumo, 10)])

    mobj = re.search(r"AO basis - number of functions:\s+(\d+)\s+number of shells:\s+(\d+)", outtext, re.MULTILINE)
    if mobj:
        psivar["N MOLECULAR ORBITALS"] = mobj.group(1)
        psivar["N BASIS FUNCTIONS"] = mobj.group(1)

    # Search for Center of charge
    mobj = re.search(
        r"Center of charge \(in au\) is the expansion point"
        + r"\n"
        + r"\s+"
        + r"X\s+=\s+([+-]?\d+[.]\d+)"
        + r"\s+"
        + r"Y\s+=\s+([+-]?\d+[.]\d+)"
        + r"\s+"
        + r"Z\s+=\s+([+-]?\d+[.]\d+)",
        outtext,
        re.MULTILINE,
    )
    if mobj:
        psivar["CENTER OF CHARGE"] = np.array([mobj.group(1), mobj.group(2), mobj.group(3)])

    mobj = re.search(
        r"Dipole moment"
        + r".*?"
        + r"A\.U\."
        + r"\s+"
        + r"DMX\s+([+-]?\d+[.]\d+)\s+"
        + r"DMXEFC\s+[+-]?\d+[.]\d+\s+"
        + r"DMY\s+([+-]?\d+[.]\d+)\s+"
        + r"DMYEFC\s+[+-]?\d+[.]\d+\s+"
        + r"DMZ\s+([+-]?\d+[.]\d+)\s+"
        + r"DMZEFC\s+[+-]?\d+[.]\d+\s+"
        + r"\-EFC\-"
        + r".*?"
        + r"A\.U\.\s+"
        + r"Total dipole\s+([+-]?\d+[.]\d+\s+)",
        outtext,
        re.MULTILINE,
    )
    # + r"DMY\s+" + r"([+-]?\d+[.]\d+)", outtext, re.MULTILINE)
    if mobj:
        psivar["DIPOLE MOMENT"] = np.array([mobj.group(1), mobj.group(2), mobj.group(3)])
        psivar["TOTAL DIPOLE MOMENT"] = mobj.group(4)
    # Process CURRENT energies (TODO: needs better way)
    if "HF TOTAL ENERGY" in psivar:
        psivar["SCF TOTAL ENERGY"] = psivar["HF TOTAL ENERGY"]
        psivar["CURRENT REFERENCE ENERGY"] = psivar["HF TOTAL ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["HF TOTAL ENERGY"]

    if "MCSCF TOTAL ENERGY" in psivar:
        psivar["CURRENT REFERENCE ENERGY"] = psivar["MCSCF TOTAL ENERGY"]
        psivar["CURRENT CORRELATION ENERGY"] = 0.0
        psivar["CURRENT ENERGY"] = psivar["MCSCF TOTAL ENERGY"]

    if "MP2 TOTAL ENERGY" in psivar and "MP2 CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["MP2 CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["MP2 TOTAL ENERGY"]

    if "MP3 TOTAL ENERGY" in psivar and "MP3 CORRECTION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["MP3 TOTAL ENERGY"] - psivar["HF TOTAL ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["MP3 TOTAL ENERGY"]

    if "MP4 TOTAL ENERGY" in psivar and "MP4 CORRECTION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["MP4 TOTAL ENERGY"] - psivar["HF TOTAL ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["MP4 TOTAL ENERGY"]

    if "CISD TOTAL ENERGY" in psivar and "CISD CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CISD CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CISD TOTAL ENERGY"]

    if "QCISD TOTAL ENERGY" in psivar and "QCISD CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["QCISD CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["QCISD TOTAL ENERGY"]

    if "LCCD TOTAL ENERGY" in psivar and "LCCD CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["LCCD CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["LCCD TOTAL ENERGY"]

    if "LCCSD TOTAL ENERGY" in psivar and "LCCSD CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["LCCSD CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["LCCSD TOTAL ENERGY"]

    if "DFT TOTAL ENERGY" in psivar:
        if "CURRENT REFERENCE ENERGY" not in psivar:
            # already set for DH-DFT
            psivar["CURRENT REFERENCE ENERGY"] = psivar["DFT TOTAL ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["DFT TOTAL ENERGY"]

    # Process TCE CURRENT energies
    # Need to be fixed
    # HOW TO KNOW options['NWCHEM']['NWCHEM_TCE']['value']?
    # TODO: CURRENT ENERGY = TCE ENERGY
    if "%s TOTAL ENERGY" % (cc_name) in psivar and ("%s CORRELATION ENERGY" % (cc_name) in psivar):
        psivar["CURRENT CORRELATION ENERGY"] = psivar["%s CORRELATION ENERGY" % (cc_name)]
        psivar["CURRENT ENERGY"] = psivar["%s TOTAL ENERGY" % (cc_name)]

    if "CCD TOTAL ENERGY" in psivar and "CCD CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CCD CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CCD TOTAL ENERGY"]

    if "CC2 TOTAL ENERGY" in psivar and "CC2 CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CC2 CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CC2 TOTAL ENERGY"]

    if "CCSD TOTAL ENERGY" in psivar and "CCSD CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CCSD CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CCSD TOTAL ENERGY"]

    if "CCSD+T(CCSD) TOTAL ENERGY" in psivar and "CCSD+T(CCSD) CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CCSD+T(CCSD) CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CCSD+T(CCSD) TOTAL ENERGY"]

    if "CCSD(T) TOTAL ENERGY" in psivar and "CCSD(T) CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CCSD(T) CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CCSD(T) TOTAL ENERGY"]

    if "CCSDT TOTAL ENERGY" in psivar and "CCSDT CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CCSDT CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CCSDT TOTAL ENERGY"]

    if ("EOM-%s TOTAL ENERGY" % (cc_name) in psivar) and ("%s EXCITATION ENERGY" % (cc_name) in psivar):
        psivar["CURRENT ENERGY"] = psivar["EOM-%s TOTAL ENERGY" % (cc_name)]
        psivar["CURRENT EXCITATION ENERGY"] = psivar["%s EXCITATION ENERGY" % (cc_name)]

    psivar[f"N ATOMS"] = len(psivar_coord.symbols)

    return psivar, psivar_coord, psivar_grad, version, module, error


def harvest_hessian(hess: str) -> np.ndarray:
    """Parses the contents of the NWChem hess file into a hessian array.

    Args:
        hess (str): Contents of the hess file
    Returns:
        (np.ndarray) Hessian matrix as a 2D array
    """

    # Change the "D[+-]" notation of Fortran output to "E[+-]" used by Python
    hess_conv = hess.replace("D", "E")

    # Parse all of the float values
    hess_tri = [float(x) for x in hess_conv.strip().splitlines()]

    # The value in the Hessian matrix is the lower triangle printed row-wise (e.g., 0,0 -> 1,0 -> 1,1 -> ...)
    n = int(np.sqrt(8 * len(hess_tri) + 1) - 1) // 2  # Size of the 2D matrix

    # Add the lower diagonal
    hess_arr = np.zeros((n, n))
    hess_arr[np.tril_indices(n)] = hess_tri

    # Transpose and then set the lower diagonal again
    hess_arr = np.transpose(hess_arr)  # Numpy implementations might only change the ordering to column-major
    hess_arr[np.tril_indices(n)] = hess_tri

    return hess_arr.T  # So that the array is listed in C-order, needed by some alignment routines


def harvest(
    in_mol: "Molecule", method: str, nwout: str, **outfiles
) -> Tuple[PreservingDict, Optional[np.ndarray], list, "Molecule", str, str]:
    """Parses all the pieces of output from NWChem: the stdout in
    *nwout* Scratch files are not yet considered at this moment.

    Args:
        in_mol (Molecule): Input molecule
        nwout (str): NWChem output molecule
        outfiles (dict): Dictionary of outfile files and their contents
    Returns:
        - (PreservingDict) Variables extracted from the output file in the last complete step
        - (None): Hessian from the last complete step (Not yet implemented)
        - (list): Gradient from the last complete step
        - (Molecule): Molecule from the last complete step
        - (str): Version string
        - (str): Error message, if any
    """

    # Parse the NWChem output
    qcvars, calc_mol, calc_grad, version, module, error = harvest_output(nwout)

    # If available, read higher-accuracy gradients from rtdb commissioned by NWC Python Task
    if outfiles.get("nwchem.grad") is not None:
        logger.debug("Reading higher-accuracy gradients")
        calc_grad = json.loads(outfiles.get("nwchem.grad"))
        calc_grad = np.array(calc_grad).reshape((-1, 3))

    # If available, read the hessian
    out_hess = None
    if outfiles.get("nwchem.hess") is not None:
        out_hess = harvest_hessian(outfiles.get("nwchem.hess"))

    # Sometimes the hierarchical setting of CURRENT breaks down
    if method == "ccsd+t(ccsd)":
        qcvars["CURRENT CORRELATION ENERGY"] = qcvars["CCSD+T(CCSD) CORRELATION ENERGY"]
        qcvars["CURRENT ENERGY"] = qcvars["CCSD+T(CCSD) TOTAL ENERGY"]

    # Make sure the input and output molecules are the same
    if calc_mol:
        if in_mol:
            if abs(calc_mol.nuclear_repulsion_energy() - in_mol.nuclear_repulsion_energy()) > 1.0e-3:
                raise ValueError(
                    f"""NWChem outfile (NRE: {calc_mol.nuclear_repulsion_energy()}) inconsistent with AtomicInput.molecule (NRE: {in_mol.nuclear_repulsion_energy()})."""
                )
    else:
        raise ValueError("""No coordinate information extracted from NWChem output.""")

    # Frame considerations
    # * `in_mol` built with deliberation and with all fields accessible.
    # * `calc_mol` has the internally consistent geometry frame but otherwise dinky (geom & symbols & maybe chgmult).
    if in_mol.fix_com and in_mol.fix_orientation:
        # Impose input frame if important as signalled by fix_*=T
        # If present, align the gradients and hessian with the original molecular coordinates
        #  NWChem rotates the coordinates of the input molecule. `calc_mol` contains the coordinates for the
        #  rotated molecule, which we can use to determine how to rotate the gradients/hessian
        # Beware the loose symmetrizer that can noticably change the input geometry.
        #  `geometry__autosym = 1d-4` tightens.
        return_mol = in_mol
        _, data = calc_mol.align(in_mol, atoms_map=True, verbose=0, mols_align=0.01)
        mill = data["mill"]

    else:
        return_mol, _ = in_mol.align(calc_mol, atoms_map=True, verbose=0, mols_align=0.01)
        mill = qcel.molutil.compute_scramble(
            len(in_mol.symbols), do_resort=False, do_shift=False, do_rotate=False, do_mirror=False
        )  # identity AlignmentMill

    return_grad = None
    if calc_grad is not None:
        return_grad = mill.align_gradient(calc_grad)

    return_hess = None
    if out_hess is not None:
        return_hess = mill.align_hessian(np.array(out_hess))

    return qcvars, return_hess, return_grad, return_mol, version, module, error



"""
Utilities for the testing suite.
"""

import sys
from typing import List, Tuple

import numpy as np
import pytest
import qcelemental as qcel
from packaging.version import parse
from pydantic import ConfigDict
from qcelemental.util import which, which_import

import qcengine as qcng

QCENGINE_RECORDS_COMMIT = "19b843b"


def _check_qcenginerecords(return_data=False):

    skip = True
    try:
        import qcenginerecords

        qcer_hash = qcenginerecords.__git_revision__[:7]
        if qcer_hash != QCENGINE_RECORDS_COMMIT[:7]:
            msg = f"Incorrect QCEngineRecord Git Revision, found {qcer_hash} need {QCENGINE_RECORDS_COMMIT[:7]}."
        else:
            skip = False
            msg = "Works!"

    except ModuleNotFoundError:
        msg = "Could not find QCEngineRecords in PYTHONPATH"

    if return_data:
        return skip, msg
    else:
        return pytest.mark.skipif(skip, reason=msg)


using_qcenginerecords = _check_qcenginerecords()


def qcengine_records(program):

    skip, msg = _check_qcenginerecords(return_data=True)
    if skip:
        pytest.skip(msg, allow_module_level=True)

    import qcenginerecords

    return qcenginerecords.get_info(program)


def is_program_new_enough(program, version_feature_introduced):
    """Returns True if `program` registered in QCEngine, locatable in
    environment, has parseable version, and that version in normalized
    form is equal to or later than `version_feature_introduced`.

    """
    if program in qcng.list_all_procedures():
        try:
            harness = qcng.get_procedure(program)
        except qcng.exceptions.ResourceError:
            return False
    else:
        harness = qcng.get_program(program, check=False)
        if not harness.found():
            return False

    candidate_version = harness.get_version()

    return parse(candidate_version) >= parse(version_feature_introduced)


def is_mdi_new_enough(version_feature_introduced):
    if which_import("mdi", return_bool=True):
        import mdi

        candidate_version = ".".join(
            [str(mdi.MDI_MAJOR_VERSION), str(mdi.MDI_MINOR_VERSION), str(mdi.MDI_PATCH_VERSION)]
        )
        return parse(candidate_version) >= parse(version_feature_introduced)
    else:
        return False


@pytest.fixture(scope="function")
def failure_engine(schema_versions, request):
    unique_name = "testing_random_name"

    class FailEngine(qcng.programs.ProgramHarness):
        iter_modes: List[str] = []
        ncalls: int = 0
        start_distance: float = 5
        equilibrium_distance: float = 4

        _defaults = {
            "name": unique_name,
            "scratch": False,
            "thread_safe": True,
            "thread_parallel": False,
            "node_parallel": False,
            "managed_memory": False,
        }

        model_config = ConfigDict(
            frozen=False,
        )

        @staticmethod
        def found(raise_error: bool = False) -> bool:
            return True

        def compute(self, input_data: "AtomicInput", config: "TaskConfig") -> "AtomicResult":
            self.ncalls += 1
            mode = self.iter_modes.pop(0)

            geom = input_data.molecule.geometry
            if geom.shape[0] != 2:
                raise ValueError("Failure Test must have an input size of two.")

            grad_value = np.abs(np.linalg.norm(geom[0] - geom[1]) - self.equilibrium_distance)
            grad = [0, 0, -grad_value, 0, 0, grad_value]

            if mode == "pass":
                return qcel.models.v2.AtomicResult(
                    **{
                        "input_data": input_data,
                        "molecule": input_data.molecule,
                        "properties": {"return_energy": grad_value},
                        "return_result": grad,
                        "success": True,
                        "extras": {"ncalls": self.ncalls},
                        "provenance": {"creator": "failure_engine", "ncores": config.ncores},
                    }
                )
            elif mode == "random_error":
                raise qcng.exceptions.RandomError("Whoops!")
            elif mode == "input_error":
                raise qcng.exceptions.InputError("Whoops!")
            else:
                raise KeyError("Testing error, should not arrive here.")

        def get_job(self):
            if from_v2(request.node.name):
                json_data = {
                    "molecule": {"symbols": ["He", "He"], "geometry": [0, 0, 0, 0, 0, self.start_distance]},
                    "specification": {
                        "driver": "gradient",
                        "model": {"method": "something"},
                    },
                }
            else:
                json_data = {
                    "molecule": {"symbols": ["He", "He"], "geometry": [0, 0, 0, 0, 0, self.start_distance]},
                    "driver": "gradient",
                    "model": {"method": "something"},
                }

            return json_data

    engine = FailEngine()
    qcng.register_program(engine)

    yield engine

    qcng.unregister_program(engine.name)


# Figure out what is imported
_programs = {
    "adcc": is_program_new_enough("adcc", "0.15.7"),
    "cfour": which("xcfour", return_bool=True),
    "dftd3": which("dftd3", return_bool=True),  # mark is classic-dftd3
    "dftd3_321": is_program_new_enough("dftd3", "3.2.1"),  # mark is classic-dftd3_321
    "dftd4": which_import("dftd4", return_bool=True),
    "dftd4_350": is_program_new_enough("dftd4", "3.5.0"),
    "s-dftd3": which_import("dftd3", return_bool=True),
    "qcore": is_program_new_enough("qcore", "0.8.9"),
    "gamess": which("rungms", return_bool=True),
    "jaguar": which_import("schrodinger.application.jaguar.input", return_bool=True),
    "mctc-gcp": is_program_new_enough("mctc-gcp", "2.3.0"),
    "gcp": which("gcp", return_bool=True),  # mark is classic-gcp
    "geometric": which_import("geometric", return_bool=True),
    "berny": which_import("berny", return_bool=True),
    "mdi": is_mdi_new_enough("1.2"),
    "molpro": is_program_new_enough("molpro", "2018.1"),
    "mopac": is_program_new_enough("mopac", "2016"),
    "mp2d": which("mp2d", return_bool=True),
    "nwchem": which("nwchem", return_bool=True),
    "optking": which_import("optking", return_bool=True),
    "psi4": is_program_new_enough("psi4", "1.2"),
    # below retired as pre-2022
    # "psi4_runqcsk": is_program_new_enough("psi4", "1.4a2.dev160"),
    # "psi4_mp2qcsk": is_program_new_enough("psi4", "1.4a2.dev580"),
    # "psi4_derqcsk": is_program_new_enough("psi4", "1.5a1.dev117"),
    "qcdb": which_import("qcdb", return_bool=True),
    "qchem": is_program_new_enough("qchem", "5.1"),
    "qcmanybody": which_import("qcmanybody", return_bool=True),
    "rdkit": which_import("rdkit", return_bool=True),
    "terachem": which("terachem", return_bool=True),
    "terachem_pbs": is_program_new_enough("terachem_pbs", "0.7.2"),
    "torchani": is_program_new_enough("torchani", "0.9"),
    "torsiondrive": which_import("torsiondrive", return_bool=True),
    "turbomole": which("define", return_bool=True),
    "xtb": which_import("xtb", return_bool=True),
    "mrchem": is_program_new_enough("mrchem", "1.0.0"),
    "mace": is_program_new_enough("mace", "0.3.2"),
    "aimnet2": which_import("pyaimnet2", return_bool=True),
    "optking_v2": is_program_new_enough("optking", "0.5"),
}
_programs["openmm"] = _programs["rdkit"] and which_import(".openmm", package="simtk", return_bool=True)


def has_program(name):
    if name == "classic-dftd3":
        return _programs["dftd3"]
    elif name == "classic-dftd3_321":
        return _programs["dftd3_321"]
    elif name == "classic-gcp":
        return _programs["gcp"]
    elif name in _programs:
        return _programs[name]
    else:
        raise KeyError(f"Program {name} not registered with QCEngine testing.")


_using_cache = {}


def _using(program: str) -> None:

    if program not in _using_cache:
        import_message = f"Not detecting module {program}. Install package if necessary to enable tests."
        skip = pytest.mark.skipif(has_program(program) is False, reason=import_message)
        general = pytest.mark.addon
        particular = getattr(pytest.mark, program)

        all_marks = (skip, general, particular)
        _using_cache[program] = [_compose_decos(all_marks), all_marks]


def _compose_decos(decos):
    # thanks, https://stackoverflow.com/a/45517876
    def composition(func):
        for deco in reversed(decos):
            func = deco(func)
        return func

    return composition


def uusing(program: str):
    """Apply 3 marks: skipif program not detected, label "addon", and label program.
    This is the decorator form for whole test functions. For example:

    @uusing("psi4")
    @uusing("dftd4")
    def test_psi4_disp():

    """
    _using(program)
    return _using_cache[program][0]


def using(program: str) -> Tuple[pytest.MarkDecorator, ...]:
    """Apply 3 marks: skipif program not detected, label "addon", and label program.
    This is the inline form for parameterizations: `marks=using("nwchem")`.
    In combo, do `marks=[*using("nwchem"), pytest.mark.quick]`

    """
    _using(program)
    return _using_cache[program][1]


@pytest.fixture(scope="function", params=[None, "as_v1", "as_v2", "to_v1", "to_v2"])  # TRUNCATE FOR CI
def schema_versions(request):
    # V1V2TEST uncomment below
    if sys.version_info >= (3, 14) and request.param != "as_v2":
        pytest.skip("Only QCSchema v2-to-v2 available for Py >=3.14")

    if request.param == "as_v1":
        return qcel.models.v1, -1, qcel.models.v1
    elif request.param == "to_v2":
        return qcel.models.v1, 2, qcel.models.v2
    elif request.param == "as_v2":
        return (qcel.models.v2, -1, qcel.models.v2)
    elif request.param == "to_v1":
        return qcel.models.v2, 1, qcel.models.v1
    else:
        return qcel.models, -1, qcel.models


@pytest.fixture(scope="function", params=["as_v1", "as_v2"])
def schema_versions2(request):
    if sys.version_info >= (3, 14) and request.param != "as_v2":
        pytest.skip("Only QCSchema v2-to-v2 available for Py >=3.14")

    if request.param == "as_v1":
        return qcel.models.v1, -1, qcel.models.v1
    elif request.param == "as_v2":
        # TODO with dict-in and dict-out and models indiscriminable and defaulting to v1
        #   the as_v2 is often not reliable, so paper over it with 2 for now. return to -1 when fixed.
        return (qcel.models.v2, 2, qcel.models.v2)


@pytest.fixture(scope="function", params=[None])
def schema_versions0(request):
    if sys.version_info >= (3, 14):
        pytest.skip("Only QCSchema v2-to-v2 available for Py >=3.14")

    return qcel.models, -1, qcel.models


@pytest.fixture(scope="function", params=[None, "as_v1", "as_v2", "to_v1", "to_v2"])
def schema_versions5(request):
    # like schema_versions except not avoiding QCSchema v1 for Py 3.14

    if request.param == "as_v1":
        return qcel.models.v1, -1, qcel.models.v1
    elif request.param == "to_v2":
        return qcel.models.v1, 2, qcel.models.v2
    elif request.param == "as_v2":
        return (qcel.models.v2, -1, qcel.models.v2)
    elif request.param == "to_v1":
        return qcel.models.v2, 1, qcel.models.v1
    else:
        return qcel.models, -1, qcel.models


def checkver_and_convert(mdl, tnm, prepost, vercheck: bool = True, cast_dict_as=None):
    import json

    import pydantic

    def check_model_v1(m):
        assert isinstance(m, pydantic.v1.BaseModel), f"type({m.__class__.__name__}) = {type(m)} ⊄ v1.BaseModel (Pyd v1)"
        assert isinstance(
            m, qcel.models.v1.basemodels.ProtoModel
        ), f"type({m.__class__.__name__}) = {type(m)} ⊄ v1.ProtoModel"
        if vercheck:
            assert m.schema_version == 1, f"{m.__class__.__name__}.schema_version = {m.schema_version} != 1"

    def check_model__v1v2(m):
        assert isinstance(
            m, (pydantic.v1.BaseModel, pydantic.BaseModel)
        ), f"type({m.__class__.__name__}) = {type(m)} ⊄ v1.BaseModel (Pyd v1)"
        assert isinstance(
            m, (qcel.models.v1.basemodels.ProtoModel, qcel.models.v2.basemodels.ProtoModel)
        ), f"type({m.__class__.__name__}) = {type(m)} ⊄ v1.ProtoModel"
        if vercheck:
            assert m.schema_version == 1, f"{m.__class__.__name__}.schema_version = {m.schema_version} != 1"

    def check_model_v2(m):
        assert isinstance(m, pydantic.BaseModel), f"type({m.__class__.__name__}) = {type(m)} ⊄ BaseModel (Pyd v2)"
        assert isinstance(
            m, qcel.models.v2.basemodels.ProtoModel
        ), f"type({m.__class__.__name__}) = {type(m)} ⊄ v2.ProtoModel"
        if vercheck:
            assert m.schema_version == 2, f"{m.__class__.__name__}.schema_version = {m.schema_version} != 2"

    if prepost == "pre":
        dict_in = isinstance(mdl, dict)
        cast_smodel = cast_dict_as or "AtomicInput"
        if "as_v1" in tnm or "to_v2" in tnm or "None" in tnm:
            if dict_in:
                mdl = getattr(qcel.models.v1, cast_smodel)(**mdl)
            # V1V2TEST check_model__v1v2(mdl)
            check_model_v1(mdl)
        elif "as_v2" in tnm or "to_v1" in tnm:
            if dict_in:
                mdl = getattr(qcel.models.v2, cast_smodel)(**mdl)
            check_model_v2(mdl)
            # NOW IN COMPUTE mdl = mdl.convert_v(1)

        if dict_in:
            mdl = mdl.model_dump()

    elif prepost == "post":
        # excuse_as_v2 is only for "post" when the compute input was a dict and while dicts are not discriminable.
        #   for now these always go to v1 in programs/model.py so as_v2 returns wrongly as v1
        # follow-up: there are too many ways this can happen, so now it's forestalled by the schema_versions fixture passing 2 to as_v2
        dict_in = isinstance(mdl, dict)
        cast_smodel = cast_dict_as or "AtomicResult"
        if "as_v1" in tnm or "to_v1" in tnm or "None" in tnm:
            if dict_in:
                mdl = getattr(qcel.models.v1, cast_smodel)(**mdl)
            # V1V2TEST check_model__v1v2(mdl)
            check_model_v1(mdl)
        elif "as_v2" in tnm or "to_v2" in tnm:
            if dict_in:
                mdl = getattr(qcel.models.v2, cast_smodel)(**mdl)
            # NOW IN COMPUTE mdl = mdl.convert_v(2)
            check_model_v2(mdl)

        if dict_in:
            # imitates compute(..., return_dict=True)
            mdl = json.loads(mdl.model_dump_json())

    return mdl


def from_v2(tnm: str) -> bool:
    """Convenience test for partitioning tests expecting QCSchema v2 input."""
    return ("to_v1" in tnm) or ("as_v2" in tnm)



"""
Testing for the DFT-D3 program harness.

Most of the tests use mindless molecules for the diversity of element species
to test as much different interactions as possible.
"""

import pprint

import numpy as np
import pytest
import qcelemental as qcel

import qcengine as qcng
from qcengine.testing import checkver_and_convert, from_v2, schema_versions, uusing


@uusing("s-dftd3")
def test_dftd3_task_b97m_m01(schema_versions, request):
    import dftd3

    models, retver, _ = schema_versions

    thr = 1.0e-8

    return_result = -0.05879001214961249

    if qcel.util.parse_version(dftd3.__version__) >= qcel.util.parse_version("1.3.0"):
        atm_correction = 6.48732559e-05
        return_result -= atm_correction

    if from_v2(request.node.name):
        atomic_input = models.AtomicInput(
            molecule=models.Molecule(**qcng.get_molecule("mindless-01", return_dict=True)),
            specification={"model": {"method": "b97m"}, "driver": "energy"},
        )
    else:
        atomic_input = models.AtomicInput(
            molecule=models.Molecule(**qcng.get_molecule("mindless-01", return_dict=True)),
            model={"method": "b97m"},
            driver="energy",
        )

    atomic_input = checkver_and_convert(atomic_input, request.node.name, "pre")
    atomic_result = qcng.compute(atomic_input, "s-dftd3", return_version=retver)
    atomic_result = checkver_and_convert(atomic_result, request.node.name, "post")

    print(atomic_result.return_result)
    assert atomic_result.success
    assert pytest.approx(atomic_result.return_result, abs=thr) == return_result


@uusing("s-dftd3")
@pytest.mark.parametrize("atm", ["nil", True, False])
@pytest.mark.parametrize(
    "inp",
    [
        pytest.param({"return_result": -0.024863196328457682, "level_hint": "d3bj"}),
        pytest.param({"return_result": -0.011922614341086182, "level_hint": "d3zero"}),
        pytest.param({"return_result": -0.054219891333201896, "level_hint": "d3mbj"}),
        pytest.param({"return_result": -0.087100927341320660, "level_hint": "d3mzero"}),
        pytest.param({"return_result": -0.014849063343059512, "level_hint": "d3op"}),
    ],
    ids=["d3bj", "d3zero", "d3mbj", "d3mzero", "d3op"],
)
def test_dftd3_task_pbe_m02(inp, schema_versions, request, atm):
    models, retver, _ = schema_versions

    # return to 1.0e-8 after https://github.com/MolSSI/QCEngine/issues/370
    thr = 1.0e-7

    return_result = inp["return_result"]

    # s-dftd3 v1.3 changed the 3-body default to False to match the executable interface.
    #   Toggle is keywords["params_tweaks"]["atm"] which isn't handled by the harness.
    #   Here, demonstrating pass-through to s-dftd3 1st-class QCSchema interface. Requires repeating method from model.
    atm_correction = 8.94882569e-05
    if atm == "nil":
        import dftd3

        xtra_kw = {}
        if qcel.util.parse_version(dftd3.__version__) >= qcel.util.parse_version("1.3.0"):
            return_result -= atm_correction
    else:
        xtra_kw = {"params_tweaks": {"method": "pbe", "atm": atm}}
        if atm is False:
            return_result -= atm_correction

    spec = {
        "model": {"method": "pbe"},
        "keywords": {"level_hint": inp["level_hint"], **xtra_kw},
        "driver": "energy",
    }

    if from_v2(request.node.name):
        atomic_input = models.AtomicInput(
            molecule=models.Molecule(**qcng.get_molecule("mindless-02", return_dict=True)),
            specification=spec,
        )
    else:
        atomic_input = models.AtomicInput(
            molecule=models.Molecule(**qcng.get_molecule("mindless-02", return_dict=True)),
            **spec,
        )

    atomic_input = checkver_and_convert(atomic_input, request.node.name, "pre")
    atomic_result = qcng.compute(atomic_input, "s-dftd3", return_version=retver)
    atomic_result = checkver_and_convert(atomic_result, request.node.name, "post")

    assert atomic_result.success
    assert pytest.approx(atomic_result.return_result, abs=thr) == return_result


@uusing("s-dftd3")
def test_dftd3_task_tpss_m02(schema_versions, request):
    models, retver, _ = schema_versions

    thr = 1.0e-8

    return_result = np.array(
        [
            [-1.38459808e-04, -3.39780053e-04, +6.63854384e-06],
            [-8.49987303e-04, -1.75644402e-05, +6.92859669e-04],
            [-1.87640802e-05, +3.48459127e-04, +3.58802840e-04],
            [-1.94903751e-04, -1.27629117e-04, -6.58938185e-04],
            [+8.81075012e-05, +6.41604548e-04, +2.74678977e-04],
            [-1.64367324e-04, -2.16197141e-04, -3.84148192e-04],
            [-1.12458629e-04, -2.56329049e-04, +4.98979702e-05],
            [-1.52971321e-05, +3.02144394e-05, +1.36451189e-04],
            [+6.66230167e-04, -8.99527309e-05, +5.62299622e-04],
            [+3.09620572e-05, -1.73604390e-05, -4.31385583e-04],
            [-1.30132201e-04, -6.02965004e-04, +2.06807277e-04],
            [-6.34318034e-04, +3.32680767e-04, -5.16509587e-04],
            [-1.90415533e-04, +4.13192949e-04, -1.76537818e-04],
            [+2.65591351e-04, -1.50550857e-04, +1.39890439e-04],
            [-3.56346859e-05, -3.56851554e-04, +1.11332319e-04],
            [+1.43384741e-03, +4.09028554e-04, -3.72139482e-04],
        ]
    )

    if from_v2(request.node.name):
        atomic_input = models.AtomicInput(
            molecule=models.Molecule(**qcng.get_molecule("mindless-02", return_dict=True)),
            specification={
                "model": {"method": ""},
                "keywords": {
                    "level_hint": "d3mbj",
                    "params_tweaks": {
                        "s8": 1.76596355,
                        "a1": 0.42822303,
                        "a2": 4.54257102,
                    },
                },
                "driver": "gradient",
            },
        )
    else:
        atomic_input = models.AtomicInput(
            molecule=models.Molecule(**qcng.get_molecule("mindless-02", return_dict=True)),
            model={"method": ""},
            keywords={
                "level_hint": "d3mbj",
                "params_tweaks": {
                    "s8": 1.76596355,
                    "a1": 0.42822303,
                    "a2": 4.54257102,
                },
            },
            driver="gradient",
        )

    atomic_input = checkver_and_convert(atomic_input, request.node.name, "pre")
    atomic_result = qcng.compute(atomic_input, "s-dftd3", return_version=retver)
    atomic_result = checkver_and_convert(atomic_result, request.node.name, "post")

    assert atomic_result.success
    assert pytest.approx(atomic_result.return_result, abs=thr) == return_result


@uusing("s-dftd3")
@pytest.mark.parametrize("atm", ["nil", True, False])
def test_dftd3_task_r2scan_m03(schema_versions, request, atm):
    models, retver, _ = schema_versions

    thr = 1.0e-8

    return_result = {
        # s9 = 0.0
        False: np.array(
            [
                [-9.35558196e-06, -7.34067064e-06, -5.91238608e-05],
                [2.53304403e-05, -5.73248858e-05, 6.46059022e-05],
                [-2.85980097e-05, 1.93908141e-05, 3.67399397e-05],
                [-7.30494324e-06, -5.48095855e-05, 3.76154049e-05],
                [1.07471133e-05, 3.03306726e-05, -2.46209798e-05],
                [-6.64928435e-05, 1.44200457e-05, -3.64523447e-05],
                [-7.88327389e-06, -3.03833956e-05, -4.75808422e-05],
                [-1.90420772e-06, -4.31459317e-05, -2.02338533e-05],
                [-3.27860042e-05, 3.68088235e-05, 2.55372601e-05],
                [4.41178314e-06, -1.11489999e-05, -4.32651442e-05],
                [7.61947667e-05, 2.64655830e-05, 2.10390818e-05],
                [-3.38080082e-05, 5.29154102e-05, -1.44995850e-04],
                [-3.85124417e-05, 4.82248838e-05, 3.75602755e-05],
                [1.01670679e-05, -3.64563841e-05, 4.13102437e-05],
                [1.02281493e-04, -2.98543913e-05, 8.22232550e-05],
                [-2.48735050e-06, 4.19080117e-05, 2.96415119e-05],
            ]
        ),
        # s9 = 1.0
        True: np.array(
            [
                [-9.15721221e-06, -8.18139252e-06, -6.04628002e-05],
                [+2.63353857e-05, -5.74347326e-05, +6.43925910e-05],
                [-2.99422439e-05, +1.88531187e-05, +3.80203831e-05],
                [-1.05590494e-05, -6.00459729e-05, +3.93481945e-05],
                [+1.59600548e-05, +3.66166973e-05, -2.69939628e-05],
                [-7.21060928e-05, +1.35991320e-05, -3.71000739e-05],
                [-9.01933781e-06, -3.41989101e-05, -4.92317946e-05],
                [-2.38625512e-06, -4.42678339e-05, -1.95513968e-05],
                [-3.09663159e-05, +3.41418638e-05, +2.51926884e-05],
                [+5.60572318e-06, -1.06356845e-05, -3.91159008e-05],
                [+7.73090102e-05, +2.75681060e-05, +2.02933984e-05],
                [-3.23109274e-05, +5.39319105e-05, -1.44309772e-04],
                [-4.05785623e-05, +5.03251549e-05, +3.92193348e-05],
                [+8.46365844e-06, -3.35282588e-05, +3.76937643e-05],
                [+1.04861677e-04, -2.99444252e-05, +8.20822571e-05],
                [-1.50951292e-06, +4.32012272e-05, +3.05230899e-05],
            ]
        ),
    }

    if atm == "nil":
        import dftd3

        xtra_kw = {}
        return_result = return_result[qcel.util.parse_version(dftd3.__version__) < qcel.util.parse_version("1.3.0")]
    else:
        xtra_kw = {"params_tweaks": {"method": "r2scan", "atm": atm}}
        return_result = return_result[atm]

    spec = {
        "keywords": {"level_hint": "d3bj", **xtra_kw},
        "driver": "gradient",
        "model": {"method": "r2scan"},
    }

    if from_v2(request.node.name):
        atomic_input = models.AtomicInput(
            molecule=models.Molecule(**qcng.get_molecule("mindless-03", return_dict=True)),
            specification=spec,
        )
    else:
        atomic_input = models.AtomicInput(
            molecule=models.Molecule(**qcng.get_molecule("mindless-03", return_dict=True)),
            **spec,
        )

    atomic_input = checkver_and_convert(atomic_input, request.node.name, "pre")
    atomic_result = qcng.compute(atomic_input, "s-dftd3", return_version=retver)
    atomic_result = checkver_and_convert(atomic_result, request.node.name, "post")

    assert atomic_result.success
    assert pytest.approx(return_result, abs=thr) == atomic_result.return_result
    assert pytest.approx(atomic_result.return_result, abs=thr) == return_result


@uusing("s-dftd3")
def test_dftd3_task_unknown_method(schema_versions, request):
    models, retver, models_out = schema_versions

    if from_v2(request.node.name):
        atomic_input = models.AtomicInput(
            molecule=models.Molecule(**qcng.get_molecule("water", return_dict=True)),
            specification={
                "keywords": {"level_hint": "d3zero"},
                "model": {"method": "non-existent-method"},
                "driver": "energy",
            },
        )
    else:
        atomic_input = models.AtomicInput(
            molecule=models.Molecule(**qcng.get_molecule("water", return_dict=True)),
            keywords={"level_hint": "d3zero"},
            model={"method": "non-existent-method"},
            driver="energy",
        )
    error = models_out.ComputeError(
        error_type="input error", error_message="No entry for 'non-existent-method' present"
    )

    atomic_input = checkver_and_convert(atomic_input, request.node.name, "pre")
    atomic_result = qcng.compute(atomic_input, "s-dftd3", return_version=retver)
    atomic_result = checkver_and_convert(
        atomic_result, request.node.name, "post", vercheck=False, cast_dict_as="FailedOperation"
    )

    print(atomic_result.error)
    assert not atomic_result.success
    assert atomic_result.error == error


@uusing("s-dftd3")
def test_dftd3_task_cold_fusion(schema_versions, request):
    models, retver, models_out = schema_versions

    if from_v2(request.node.name):
        atomic_input = models.AtomicInput(
            molecule={
                "symbols": ["Li", "Li", "Li", "Li"],
                "geometry": [
                    [-1.58746019997201, +1.58746019997201, +1.58746019997201],
                    [-1.58746019997201, +1.58746019997201, +1.58746019997201],
                    [-1.58746019997201, -1.58746019997201, -1.58746019997201],
                    [+1.58746019997201, +1.58746019997201, -1.58746019997201],
                ],
                "validated": True,  # Force a nuclear fusion input, to make dftd3 fail
            },
            specification={
                "keywords": {"level_hint": "d3zero"},
                "model": {"method": "pbe"},
                "driver": "energy",
            },
        )
    else:
        atomic_input = models.AtomicInput(
            molecule={
                "symbols": ["Li", "Li", "Li", "Li"],
                "geometry": [
                    [-1.58746019997201, +1.58746019997201, +1.58746019997201],
                    [-1.58746019997201, +1.58746019997201, +1.58746019997201],
                    [-1.58746019997201, -1.58746019997201, -1.58746019997201],
                    [+1.58746019997201, +1.58746019997201, -1.58746019997201],
                ],
                "validated": True,  # Force a nuclear fusion input, to make dftd3 fail
            },
            keywords={"level_hint": "d3zero"},
            model={"method": "pbe"},
            driver="energy",
        )
    error = models_out.ComputeError(
        error_type="input error",
        error_message="Too close interatomic distances found",
    )

    atomic_input = checkver_and_convert(atomic_input, request.node.name, "pre")
    atomic_result = qcng.compute(atomic_input, "s-dftd3", return_version=retver)
    atomic_result = checkver_and_convert(atomic_result, request.node.name, "post", vercheck=False)

    pprint.pprint(atomic_result.error.model_dump())
    assert not atomic_result.success
    assert atomic_result.error == error
